'''
本模块用于图论分析

This module is used for graph theory analysis
'''
import numpy as np


def ek(G, ori, des):
    '''
    EK算法
    利用BFS遍历求解最大流问题
    
    参数
    ----
    G：邻接矩阵(非邻接点间的边权重(即流量)为inf或负数或0)
    ori：整型，起点
    des：整型，终点
    
    返回
    ----
    数类型，最大流
    
    
    Edmond-Karp Algorithm
    Solve maximum flow problem by BFS
    
    Parameters
    ----------
    G: adjacency matrix(the weight(flow) of the edge between non-adjacent nodes is inf, negative or 0)
    ori: int, origin
    des: int, destination
    
    Return
    ------
    num, maximum flow
    '''
    G = np.array(G, dtype=np.float)
    n = len(G)
    G[G < 0] = 0
    G[G == float('inf')] = 0
    max_flow = 0
    
    while True:
        visited = np.zeros(n)
        qu = [[ori, -1]]
        path = []
        visited[ori] = 1
        
        i = 0
        judge = False
        while i < len(qu):
            for j in range(n):
                if visited[j] == 0 and G[qu[i][0]][j] > 0:
                    visited[j] = 1
                    qu.append([j, i])
                
                    if j == des:            #若找到简单路径则break
                        judge = True
                        break
            
            if judge:
                break
            
            if i == len(qu)-1:              #若没有简单路径则程序结束
                return max_flow
            
            i += 1
        
        path.append(qu[-1][0])
        loc = qu[-1][1]
        while loc != -1:
            path.append(qu[loc][0])
            loc = qu[loc][1]
        
        dis = []                            #将path中各边流量存入dis
        for i in range(len(path)-1):
            dis.append(G[path[i+1], path[i]])
        
        mindis = min(dis)
        max_flow += mindis                  #更新最大流
        
        for i in range(len(path)-1):        #更新邻接矩阵
            G[path[i+1], path[i]] -= mindis
            G[path[i], path[i+1]] += mindis


def kruskal(G):
    '''
    克鲁斯卡尔算法
    生成最小生成树
    
    参数
    ----
    G：图的邻接矩阵(非邻接点间的边权重为inf或负数)
    
    返回
    ----
    2-D列表类型，每个一维元素为：[权值，起点，终点]
    
    
    Kruskal Algorithm
    Generate minimum spanning tree
    
    Parameter
    ---------
    G: adjacency matrix(the weight of the edge between non-adjacent nodes is inf or negative)
    
    Return
    ------
    2-D list, every element is [weight, origin, destination]
    '''
    G = np.array(G, dtype=np.float)
    n = len(G)
    G[G < 0] = float('inf')
    E = []
    
    # 将边信息(边长，起点，终点)存入self.E并按从小到大排序
    for i in range(n):
        for j in range(i):
            if G[i, j] != 0 and G[i, j] != float('inf'):
                E.append([G[i, j], i, j])
    E.sort()
        
    path = []
    vset = np.arange(n)
    k = 1
    j = 0
    while k < n:
        ul = E[j][1]        #第j条边的起点
        vl = E[j][2]        #第j条边的终点
        sn1 = vset[ul]
        sn2 = vset[vl]
        if sn1 != sn2:
            path.append(E[j])
            k += 1
            vset[vset == sn2] = sn1
        j += 1
    
    return path


class Floyd():
    '''
    弗洛伊德算法
    
    参数
    ----
    G：图的邻接矩阵(非邻接点间的边权重为inf或负数)
    
    属性
    ----
    A：二维列表，A[i, j]代表节点i到节点j的距离
    
    
    Floyd Algorithm
    
    Parameter
    ---------
    G: adjacency matrix(the weight of the edge between non-adjacent nodes is inf or negative)
    
    Attribute
    ---------
    A: 2-D list, A[i, j] is the distance from node i to node j
    '''
    def __init__(self, G):
        self.A = np.array(G, dtype=np.float)
        n = len(self.A)
        self.A[self.A < 0] = float('inf')
        for i in range(n):
            self.A[i, i] = 0
        
        self.__path = np.zeros((n, n), dtype=np.int)
        for i in range(n):
            for j in range(n):
                if i != j:
                    if 0 <= self.A[i, j] < float('inf'):
                        self.__path[i, j] = i
                    else:
                        self.__path[i, j] = -1
                else:
                    self.__path[i, j] = -1
        
        for k in range(n):
            for i in range(n):
                for j in range(n):
                    if self.A[i, j] > self.A[i, k] + self.A[k, j]:
                        self.A[i, j] = self.A[i, k] + self.A[k, j]
                        self.__path[i, j] = self.__path[k, j]
    
    
    def path(self, ori, des):
        '''
        求解两节点之间的最短路径
        
        参数
        ----
        ori：整型，起点
        des：整型，终点
        
        返回
        ----
        元组类型，(最短路径的长度, 最短路径(list))
        
        
        Find the shortest path between two nodes
        
        Parameters
        ----------
        ori: int, origin
        des: int, destination
        
        Return
        ------
        tuple, (the length of the path, path(list))
        '''
        if self.A[ori, des] == float('inf'):
            raise Exception('Unable to reach the destination from the origin.')
        
        elif ori == des:
            raise Exception('The origin and the destination should be different.' )
        
        else:
            distance = self.A[ori, des]
            
            path = [des]
            while path[-1] != ori:
                path.append(self.__path[ori, des])
                des = self.__path[ori, des]
            path.reverse()
            
            return distance, path
    
    
    def circle(self, point):
        '''
        求解某一节点的最短有向环
        
        参数
        ----
        point：整型或浮点数类型，节点
        
        返回
        ----
        元组类型，(最短环的长度, 最短环(list))
        
        
        Find the shortest directed circle of a node
        
        Parameter
        ---------
        point: int or float, node
        
        Return
        ------
        tuple, (the length of the circle, circle(list))
        '''
        circle_distance_list = self.A[point] + self.A[:, point]
        circle_distance_list[point] = float('inf')
        target = np.argmin(circle_distance_list)
        distance = np.min(circle_distance_list)
        if distance == float('inf'):
            raise Exception("There is no proper circle.")
        
        target_1 = target
        path1 = [target_1]
        while path1[-1] != point:
            path1.append(self.__path[point, target_1])
            target_1 = self.__path[point, target_1]
        path1.reverse()
        
        target_2 = point
        path2 = [target_2]
        while path2[-1] != target:
            path2.append(self.__path[target, target_2])
            target_2 = self.__path[target, target_2]
        path2.reverse()
        path2.remove(path2[0])
        
        return distance, path1+path2


class Dijkstra():
    '''
    狄克斯特拉算法
    
    参数
    ----
    G：图的邻接矩阵(非邻接点间的边权重为inf或负数)
    ori：整型，起点
    
    属性
    ----
    dis：一维列表，dis[i]表示起点到节点i的距离
    
    
    Dijkstra Algorithm
    
    Parameters
    ----------
    G: adjacency matrix(the weight of the edge between non-adjacent nodes is inf or negative)
    ori: int, origin
    
    Attribute
    ---------
    dis: 1-D list, dis[i] is the distance from the origin to node i
    '''
    def __init__(self, G, ori):
        self.__ori = ori
        G = np.array(G, dtype=np.float)
        G[G < 0] = float('inf')
        n = len(G)
        self.dis = G[ori].copy()
        S = np.zeros(n)
        self.__path = self.dis.copy()
        self.__path[self.__path < float('inf')] = ori
        self.__path[self.__path == float('inf')] = -1
        self.__path = np.array(self.__path, dtype=np.int)
        
        S[ori] = 1
        self.__path[ori] = 0
        for i in range(n-1):
            mindis = float('inf')
            for j in range(n):
                if S[j] == 0 and self.dis[j] < mindis:
                    u = j
                    mindis = self.dis[j]
            S[u] = 1
            for j in range(n):
                if S[j] == 0:
                    if G[u, j] < float('inf') and self.dis[u] + G[u, j] < self.dis[j]:
                        self.dis[j] = self.dis[u] + G[u, j]
                        self.__path[j] = u
    
    
    def path(self, des):
        '''
        求解ori和des之间的最短路径
        
        参数
        ----
        des：整型，终点
        
        返回
        ----
        元组类型，(最短路径的长度, 最短路径(list))
        
        
        Find the shortest path from ori to des
        
        Parameter
        ---------
        des: int, destination
        
        Return
        ------
        tuple, (the length of the path, path(list))
        '''
        if self.__path[des] == -1:
            raise Exception('Unable to reach the destination from the origin.')
        
        elif self.__ori == des:
            raise Exception('The origin and the destination should be different.' )
        
        else:
            distance = self.dis[des]
            
            path = [des]
            while path[-1] != self.__ori:
                path.append(self.__path[des])
                des = self.__path[des]
            path.reverse()
            
            return distance, path


class Markov():
    '''
    马尔科夫链
    
    参数
    ----
    P：2-D ndarray，转移矩阵
    
    
    Markov Chain
    
    Parameter
    ---------
    P: 2-D ndarray, transition matrix
    '''
    def __init__(self, P):
        self.P = np.array(P)
    
    
    def k(self, k):
        '''
        参数
        ----
        k：整型，步长
        
        返回
        ----
        2-D ndarray，k步转移矩阵
        
        
        Parameter
        ---------
        k: int, length
        
        Return
        ------
        2-D ndarray, K-step transition matrix
        '''
        P0 = self.P
        P = P0
        for i in range(k-1):
            P = np.dot(P, P0)
        return P
    
    
    def w(self):
        '''
        稳态分布
        
        返回
        ----
        1-D ndarray，稳态分布
        
        
        Steady State Distribution
        
        Return
        ------
        1-D ndarray, steady state distribution
        '''
        A = self.P.T - np.identity(self.P.shape[0])
        A = np.vstack((A, np.ones((1, A.shape[1]))))
        b = np.zeros((self.P.shape[0] + 1, 1))
        b[-1] = 1
        Ai = np.linalg.pinv(A)
        w = np.dot(Ai, b).T
        w = w[0]
        return w


'''
本模块用于数据预处理

This module is used for data preproccessing
'''
import numpy as np
from maysics.utils import e_distances
from matplotlib import pyplot as plt
plt.rcParams['font.sans-serif'] = ['FangSong']
plt.rcParams['axes.unicode_minus'] = False
from io import BytesIO
from lxml import etree
import base64
import math


def _rc(*arg):
    cov_mat = np.cov(arg)
    var_mat = np.diagonal(cov_mat)**0.5
    var_mat[var_mat == 0] = 1
    
    for i in range(cov_mat.shape[0]):
        cov_mat[i] /= var_mat[i]
        cov_mat[:, i] /= var_mat[i]
    
    return cov_mat


def _preview_process(data, value_round):
    '''
    预览处理
    '''
    data = np.array(data, dtype=np.float)
    
    name_list = ['平均值', '中位数', '方差', '标准差', '最大值', '最小值', '偏度', '峰度']
    value_list = []
    mean_ = data.mean(axis=0)
    value_list.append(np.round(mean_, value_round))
    value_list.append(np.round(np.median(data, axis=0), value_round))
    value_list.append(np.round(data.var(axis=0), value_round))
    value_list.append(np.round(data.std(axis=0), value_round))
    value_list.append(np.round(data.max(axis=0), value_round))
    value_list.append(np.round(data.min(axis=0), value_round))
    value_list.append(np.round(((data - mean_)**3).mean(axis=0), value_round))
    value_list.append(np.round(((data - mean_)**4).mean(axis=0), value_round))
    value_list = np.array(value_list).flatten()
    
    style = '''
    <style>
    table{
        border-collapse: collapse;
    }
    table, table tr td {
        border:1px solid #ccc;
    }
    table tr td{
        padding: 5px 10px;
    }
    </style>
    '''
    table = '<h2 style="padding-left:50px; border-top:1px solid #ccc">数值特征</h2>' + style + '<table align="center"><caption></caption>'
    for i in range(8):
        table += '<tr><td>' + name_list[i] + '</td>' + '<td>%s</td>' * data.shape[1] + '</tr>'
    table = '<h1 style="padding-left:50px;">数据信息</h1>' + table % tuple(value_list) + '</table>'
    
    
    data = data.T
    num = data.shape[0]
    plt.figure(figsize=(9, 3 * num))
    for i in range(num):
        q1, q2, q3 = np.percentile(data[i], [25, 50, 75])
        plt.scatter(mean_[i], i+1, marker='o', color='white', s=30, zorder=3)
        plt.hlines(i+1, q1, q3, color='k', linestyle='-', lw=1)
    ax = plt.violinplot(data.tolist(), showextrema=False, vert=False)
    plt.title('分布图')
    
    buffer = BytesIO()
    plt.savefig(buffer)
    plot_data = buffer.getvalue()
    imb = base64.b64encode(plot_data)
    ims = imb.decode()
    imd = 'data:image/png;base64,' + ims
    im1 = '<div align="center"><img src="%s"></div>' % imd
    im1 = '<br></br><h2 style="padding-left:50px; border-top:1px solid #ccc">密度分布</h2>' + im1
    
    cov_mat = _rc(*data)
    matrix = '<table border="0"><caption></caption>'
    
    for i in range(num):
        matrix += '<tr>' + '<td>%s</td>' * num + '</tr>'
    matrix = matrix % tuple(np.round(cov_mat.flatten(), value_round)) + '</table>'
    
    
    plt.figure(figsize=(8, 8))
    plt.matshow(cov_mat, fignum=0, cmap='Blues')
    plt.colorbar()
    plt.title('相关系数图')
    
    buffer = BytesIO()
    plt.savefig(buffer)
    plot_data = buffer.getvalue()
    imb = base64.b64encode(plot_data)
    ims = imb.decode()
    imd = 'data:image/png;base64,' + ims
    im2 = '<div style="display:flex;flex-direction:row;vertical-align:middle;justify-content:center;width:100%;height:80vh"><div style="margin:auto 0;white-space:pre-wrap;max-width:50%">'
    im2 = im2 +'相关矩阵：'+ matrix + '</div><img style="object-fit:contain;max-width:45%;max-height:80vh" src="{}"/></div>'.format(imd)
    im2 = '<br></br><h2 style="padding-left:50px; border-top:1px solid #ccc">相关性</h2>' + im2
    
    
    plt.figure(figsize=(2.5 * num, 2.5 * num))
    for i in range(num * num):
        ax = plt.subplot(num, num, i+1)
        ax.plot(data[i//num], data[i%num], 'o')
    
    buffer = BytesIO()
    plt.savefig(buffer)
    plot_data = buffer.getvalue()
    imb = base64.b64encode(plot_data)
    ims = imb.decode()
    imd = "data:image/png;base64," + ims
    im3 = '<div align="center"><img src="%s"></div>' % imd
    im3 = '<br></br><h2 style="padding-left:50px; border-top:1px solid #ccc">散点关系</h2>' + im3
    
    return '<title>数据信息预览</title>' + table + im1 + im2 + im3


def preview_file(filename, data, value_round=3):
    '''
    生成数据预览报告的html文件
    
    参数
    ----
    filename：字符串类型，文件名
    data：二维数组，数据
    value_round：整型，数字特征保留的小数点后的位数
    
    
    Generate preview report with html file
    
    Parameters
    ----------
    filename: str, file name
    data: 2-D array, data
    value_round: int, the number of digits after the decimal point retained by numeric features
    '''
    root = _preview_process(data=data, value_round=value_round)
    html = etree.HTML(root)
    tree = etree.ElementTree(html)
    tree.write(filename)


def preview(data, value_round=3):
    '''
    在jupyter中显示数据预览报告
    
    参数
    ----
    data：二维数组，数据
    value_round：整型，数字特征保留的小数点后的位数
    
    
    Display preview report in jupyter
    
    Parameters
    ----------
    data: 2-D array, data
    value_round: int, the number of digits after the decimal point retained by numeric features
    '''
    root = _preview_process(data=data, value_round=value_round)
    
    from IPython.core.display import display, HTML
    display(HTML(root))


def length_pad(seq, maxlen=None, value=0, padding='pre', dtype='int32'):
    '''
    填充二维列表，使得每行长度都为maxlen
    
    参数
    ----
    seq：二维列表，需要填充的对象
    maxlen：整型，可选，每行的最大长度，默认为原二维列表最大的长度
    value：数类型，可选，填充值，默认为0
    padding：字符串类型，可选，填充位置，'pre'代表从前面填充，'post'代表从后面填充，默认为'pre'
    dtype：可选，输出的元素类型，默认为'int32'
    
    返回
    ----
    二维ndarray
    
    
    Pad the 2-D list so that every row is 'maxlen' in length
    
    Parameters
    ----------
    seq: 2-D list, objects that need to be padded
    maxlen: int, callable, the maximum length of each row, default = the maximum length of the original 2-D list
    value: num, callable, padding value, default=0
    padding: str, callable, padding location, 'pre' means padding from the front and 'post' from the back, default='pre'
    dtype: callable, the element type of the output, default='int32'
    
    Return
    ------
    2-D ndarray
    '''
    seq = list(seq)
    if not maxlen:
        maxlen = 0
        for i in seq:
            if len(i) > maxlen:
                maxlen = len(i)
    
    if padding == 'pre':
        for i in range(len(seq)):
            if maxlen > len(seq[i]):
                seq[i] = [value] * (maxlen - len(seq[i])) + seq[i]
            elif maxlen < len(seq[i]):
                seq[i] = seq[i][-1 * maxlen:]
    
    elif padding == 'post':
        for i in range(len(seq)):
            if maxlen > len(seq[i]):
                seq[i] += [value] * (maxlen - len(seq[i]))
            elif maxlen < len(seq[i]):
                seq[i] = seq[i][:maxlen]
    
    return np.array(seq, dtype=dtype)


def sample_pad(data, index=0, padding=None):
    '''
    对二维数据进行样本填充
    先对data中的每个二维数据进行遍历，以各个index列的值作为全集，再对data的每个二维数据进行填充
    如：data1 = [[0, 1],
                 [1, 2],
                 [2, 3]]
        data2 = [[2, 3],
                 [3, 4],
                 [4, 5]]
        data = (data1, data2)
        则得到输出：
        output = [array([[0, 1],
                        [1, 2],
                        [2, 3],
                        [3, nan],
                        [4, nan]]),
                  
                  array([[0, nan],
                         [1,nan],
                         [2, 3],
                         [3, 4],
                         [4, 5]])]
    
    data：元组或列表类型，数据
    index：整型，作为扩充全集的标准列的索引
    padding：填充值，可选，默认为None
    
    
    Sample filling for 2D data
    Values of each index column will be taken as the complete set, then each two-dimensional data of data is padded
    e.g. data1 = [[0, 1],
                 [1, 2],
                 [2, 3]]
         data2 = [[2, 3],
                  [3, 4],
                  [4, 5]]
         data = (data1, data2)
         output = [array([[0, 1],
                          [1, 2],
                          [2, 3],
                          [3, nan],
                          [4, nan]]),
                    
                    array([[0, nan],
                           [1,nan],
                           [2, 3],
                           [3, 4],
                           [4, 5]])]
    
    data: tuple or list, data
    index: int, the index of a standard column as an extended complete set
    padding: padding value, optional, default=None
    '''
    time_set = set()
    result = []
    if not padding:
        padding = [np.nan] * (len(data[0][0]) - 1)
    else:
        padding = list(padding)
    
    for i in range(len(data)):
        data_part = np.array(data[i], dtype=np.object)
        result.append(data_part)
        time_set = time_set | set(data_part[:, index])
    
    for i in range(len(result)):
        different_set_list = np.array([list(time_set - set(result[i][:, index]))], dtype=np.object).T
        num = len(different_set_list)
        padding_new = np.array(padding * num, dtype=np.object).reshape(num, -1)
        different_set_list = np.hstack((different_set_list, padding_new))
        result[i] = np.vstack((result[i], different_set_list))
    
    return result


def shuffle(*arg):
    '''
    以相同方法打乱多个序列或打乱一个序列
    
    返回
    ----
    一个ndarray
    
    
    shuffle multiple sequences in the same way or shuffle a sequences
    
    Return
    ------
    a ndarray
    '''
    state = np.random.get_state()
    a_new_list = []
    for li in arg:
        np.random.set_state(state)
        np.random.shuffle(li)
        a_new_list.append(li)
    return np.array(a_new_list)


def data_split(data, targets, train_size=None, test_size=None, shuffle=True, random_state=None):
    '''
    分离数据
    
    参数
    ----
    data：数据
    targets：指标
    train_size：浮点数类型，可选，训练集占总数据量的比，取值范围为(0, 1]，默认为0.75
    test_size：浮点数类型，可选，测试集占总数据量的比，取值范围为[0, 1)，当train_size被定义时，该参数无效
    shuffle：布尔类型，可选，True表示打乱数据，False表示不打乱数据，默认为True
    random_state：整型，可选，随机种子
    
    返回
    ----
    元组，(数据测试集, 指标测试集, 数据验证集, 指标验证集)
    
    
    split the data
    
    Parameters
    ----------
    data: data
    targets: targets
    train_size: float, callable, ratio of training set to total data, value range is (0, 1], default=0.75
    test_size: float, callable, ratio of test set to total data, value range is [0, 1)
    shuffle: bool, callable, 'True' will shuffle the data, 'False' will not, default = True
    random_state: int, callable, random seed
    
    Return
    ------
    tuple, (train_data, train_target, validation_data, validation_target)
    '''
    if not (train_size or test_size):
        train_size = 0.75
    elif test_size:
        train_size = 1 - test_size
    
    if train_size <= 0 or train_size > 1:
        raise Exception("'train_size' should be in (0, 1], 'test_size' should be in [0, 1)")
    
    if shuffle:
        np.random.seed(random_state)
        
        state = np.random.get_state()
        np.random.shuffle(data)
        np.random.set_state(state)
        np.random.shuffle(targets)
        
    num_of_data = len(data)
    train_data = data[:int(num_of_data * train_size)]
    train_target = targets[:int(num_of_data * train_size)]
    validation_data = data[int(num_of_data * train_size):]
    validation_target = targets[int(num_of_data * train_size):]
    
    return train_data, train_target, validation_data, validation_target


def dataloader(data, targets, choose_rate=0.3, is_shuffle=True, random_state=None):
    '''
    数据随机生成器
    
    参数
    ----
    data：数据
    targets：指标
    choose_rate：浮点数类型，可选，生成率，范围为[0, 1]，默认为0.3
    is_shuffle：布尔类型，可选，True表示打乱数据，False表示不打乱数据，默认为True
    random_state：整型，可选，随机种子
    
    返回
    ----
    生成器
    
    
    Data Random Generator
    
    Parameters
    ----------
    data: data
    targets: targets
    choose_rate: float, callable, generation rate whose range is [0, 1], default=0.3
    is_shuffle: bool, callable, 'True' will shuffle the data, 'False' will not, default = True
    random_state: int, callable, random seed
    
    Return
    ------
    generator
    '''
    data = np.array(data)
    targets = np.array(targets)
    
    if is_shuffle:
        np.random.seed(random_state)
        data, targets = shuffle(data, targets)
    num = len(data)
    choose_rate = int(num * choose_rate)
    times = int(math.ceil(num / choose_rate))
    
    for i in range(times):
        loc_1 = i * choose_rate
        loc_2 = (i + 1) * choose_rate
        yield data[loc_1: loc_2], targets[loc_1: loc_2]


def kfold(data, targets, n, k=5):
    '''
    参数
    ----
    data：数据
    targets：指标
    n：整型，表示将第n折作为验证集，从0开始
    k：整型，可选，k折验证的折叠数，默认k=5
    
    返回
    ----
    元组，(数据测试集, 指标测试集, 数据验证集, 指标验证集)
    
    
    Parameters
    ----------
    data: data
    targets: targets
    n: int, take the nth part as validation set, starting from 0
    k: int, callable, the number of k-fold, default = 5
    
    Return
    ------
    tuple, (train_data, train_target, validation_data, validation_target)
    '''
    num_validation_samples = len(data) // k
    
    validation_data = data[num_validation_samples * n:
                           num_validation_samples * (n + 1)]
    validation_targets = targets[num_validation_samples * n:
                                 num_validation_samples * (n + 1)]
    
    train_data = np.concatenate([data[: num_validation_samples * n],
                                 data[num_validation_samples * (n + 1):]])
    train_targets = np.concatenate([targets[: num_validation_samples * n],
                                    targets[num_validation_samples * (n + 1):]])
    
    return train_data, train_targets, validation_data, validation_targets


def standard(data, index=None, mean=True, var=True):
    '''
    标准化数据
    z = (x - u) / s
    z：新数据；  x：原数据；  u：均值；  s：方差
    如果某一列数据完全相同（即方差s=0），则该列数据全部归零
    
    参数
    ----
    data：2-D的ndarray数据
    index：列表形式，可选，需要进行标准化的列的索引，默认为全部
    mean：布尔类型，可选，是否将均值调整为0
    var：布尔类型，可选，是否将方差调整为1
    
    返回
    ----
    2-D ndarray
    
    
    Standardize data
    z = (x - u) / s
    z: new data;  x: origin data;  u: mean value;  s: variance
    if data in one column are the same(s=0), data in this column will be turned to 0
    
    Parameters
    ----------
    data: 2-D ndarray
    index: list, callable, index of columns need to be standardized, defalut to all
    mean: bool, callable, if adjust the mean value to 0
    var: bool, callable, if adjust the variance to 0
    
    Return
    ------
    2-D ndarray
    '''
    data=np.array(data, dtype=np.float)
    
    if index:
        if mean:
            mean = data[:, index].mean(axis=0)
        else:
            mean = 0.0
        data[:, index] -= mean
        
        if var:
            std = data[:, index].std(axis=0)
            std_zero_indices = np.nonzero(std == 0)
            std[std==0] = 1.0
            data[:, index] /= std
            if list(std_zero_indices[0]):
                for i in std_zero_indices[0]:
                    data[:, index][:, i] *= 0
    else:
        if mean:
            mean = data.mean(axis=0)
        else:
            mean = 0.0
        data -= mean
        
        if var:
            std = data.std(axis=0)
            std_zero_indices = np.nonzero(std == 0)
            std[std==0] = 1.0
            data /= std
            if list(std_zero_indices[0]):
                for i in std_zero_indices[0]:
                    data[:, i] *= 0
    
    return data


def minmax(data, index=None, feature_range=(0, 1)):
    '''
    归一化数据
    X_std = (X - X.min(axis=0)) / (X.max(axis=0) - X.min(axis=0))
    X_new = X_std * (feature_range[1] - feature_range[0]) + feature_range[0]
    
    参数
    ----
    data：2-D的ndarray数据
    index：列表形式，可选，需要进行标准化的列的索引，默认为全部
    feature_range：元组形式，可选，需要转换的范围，默认为(0, 1)
    
    返回
    ----
    2-D ndarray
    
    
    Normalized data
    X_std = (X - X.min(axis=0)) / (X.max(axis=0) - X.min(axis=0))
    X_new = X_std * (feature_range[1] - feature_range[0]) + feature_range[0]
    
    Parameters
    ----------
    data: 2-D的ndarray数据
    index: list, callable, index of columns need to be standardized, defalut to all
    feature_range: tuple, callabel, final range of transformed data
    
    Return
    ------
    2-D ndarray
    '''
    data=np.array(data, dtype=np.float)
    
    if index:
        min_data = data[:, index].min(axis=0)
        length = data[:, index].max(axis=0) - min_data
        data[:, index] = (data[:, index] - min_data) / length
        data[:, index] = data[:, index] * (feature_range[1] - feature_range[0]) + feature_range[0]
    
    else:
        min_data = data.min(axis=0)
        length = data.max(axis=0) - min_data
        data = (data - min_data) / length
        data = data * (feature_range[1] - feature_range[0]) + feature_range[0]
    
    return data


def normalizer(data, index=None):
    '''
    使每个数据的模为1
    
    参数
    ----
    data：2-D的ndarray数据
    index：列表形式，可选，需要进行标准化的列的索引，默认为全部
    
    返回
    ----
    2-D ndarray
    
    
    Making the moduli of data equal 1
    
    Parameters
    ----------
    data: 2-D的ndarray数据
    index: list, callable, index of columns need to be standardized, defalut to all
    
    Return
    ------
    2-D ndarray
    '''
    data = np.array(data, dtype=np.float)
    
    if index:
        distance_list = e_distances(data[:, index])
        distance_list[distance_list == 0] = 1
        print(distance_list)
        data[:, index] /= distance_list
    
    else:
        distance_list = e_distances(data)
        distance_list[distance_list == 0] = 1
        print(distance_list)
        data /= distance_list
    
    return data


class Standard():
    '''
    标准化数据
    以一组数据为标准，将其他数据按照该组数据的方差和均值进行标准化
    z = (x - u) / s
    z：新数据；  x：原数据；  u：均值；  s：方差
    如果某一列数据完全相同（即方差s=0），则该列数据全部归零
    
    参数
    ----
    data：2-D的ndarray数据
    mean：布尔类型，可选，是否将均值调整为0
    var：布尔类型，可选，是否将方差调整为1
    
    属性
    ----
    mean：1-D的ndarray，数据的均值
    var：1-D的ndarray，数据的方差
    
    
    Standardize data
    Take a group of data as the standard, other data will be standardized according to the variance and mean value of this group of data
    z = (x - u) / s
    z: new data;  x: origin data;  u: mean value;  s: variance
    if data in one column are the same(s=0), data in this column will be turned to 0
    
    Parameters
    ----------
    data: 2-D ndarray
    mean: bool, callable, if adjust the mean value to 0
    var: bool, callable, if adjust the variance to 0
    
    Attributes
    ----------
    mean: 1-D ndarray, mean values of the data
    var: 1-D ndarray, variances of the data
    '''
    def __init__(self, data, mean=True, var=True):
        data=np.array(data, dtype=np.float)
        
        if mean:
            self.mean = data.mean(axis=0)
        else:
            self.mean = False
        
        if var:
            self.var = data.var(axis=0)
            std_zero_indices = np.nonzero(self.var == 0)
            self.var[self.var==0] = 1.0
            self.__std_zero_indices = std_zero_indices[0]
        else:
            self.var = False
    
    
    def transform(self, data, index=None):
        '''
        参数
        ----
        data：2-D的ndarray数据
        index：列表形式，可选，需要进行标准化的列的索引，默认为全部
        
        返回
        ----
        2-D ndarray
        
        
        Parameters
        ----------
        data: 2-D ndarray
        index: list, callable, index of columns need to be standardized, defalut to all
        
        Return
        ------
        2-D ndarray
        '''
        data=np.array(data, dtype=np.float)
        if index:
            if type(self.mean).__name__ == 'ndarray':
                data[:, index] -= self.mean[index]
            
            if type(self.var).__name__ == 'ndarray':
                data[:, index] /= self.var[index]
        
        else:
            if type(self.mean).__name__ == 'ndarray':
                data -= self.mean
            
            if type(self.var).__name__ == 'ndarray':
                data /= self.var
        
        return data


class Minmax():
    '''
    归一化数据
    以一组数据为标准，将其他数据按照该组数据的最大值和最小值进行归一化
    X_std = (X - X.min(axis=0)) / (X.max(axis=0) - X.min(axis=0))
    X_new = X_std * (feature_range[1] - feature_range[0]) + feature_range[0]
    
    参数
    ----
    data：2-D的ndarray数据
    
    属性
    ----
    minimum：1-D ndarray，最小值
    maximum：1-D ndarray，最大值
    length：1-D ndarray，最大值 - 最小值
    
    
    Normalized data
    Take a group of data as the standard, other data will be normalized according to the minimum and maximum value of this group of data
    X_std = (X - X.min(axis=0)) / (X.max(axis=0) - X.min(axis=0))
    X_new = X_std * (feature_range[1] - feature_range[0]) + feature_range[0]
    
    Parameter
    ---------
    data: 2-D的ndarray数据
    
    Attributes
    ----------
    minimum: 1-D ndarray
    maximum: 1-D ndarray
    length: 1-D ndarray, maximum - minimum
    '''
    def __init__(self, data):
        data=np.array(data, dtype=np.float)
        
        self.minimum = data.min(axis=0)
        self.maximum = data.max(axis=0)
        self.length = self.maximum - self.minimum
    
    
    def transform(self, data, index=None, feature_range=(0, 1)):
        '''
        参数
        ----
        data：2-D的ndarray数据
        index：列表形式，可选，需要进行标准化的列的索引，默认为全部
        feature_range：元组形式，可选，需要转换的范围，默认为(0, 1)
        
        返回
        ----
        2-D ndarray
        
        
        Parameters
        ----------
        data: 2-D的ndarray数据
        index: list, callable, index of columns need to be standardized, defalut to all
        feature_range: tuple, callabel, final range of transformed data
        
        
        Return
        ------
        2-D ndarray
        '''
        data=np.array(data, dtype=np.float)
        if index:
            data[:, index] = (data[:, index] - self.minimum[index]) / self.length[index]
            data[:, index] = data[:, index] * (feature_range[1] - feature_range[0]) + feature_range[0]
        
        else:
            data = (data - self.minimum) / self.length
            data = data * (feature_range[1] - feature_range[0]) + feature_range[0]
        
        return data


class RC():
    '''
    相关系数
    
    参数
    ----
    *arg：列表类型
    
    属性
    ----
    rc_mat：相关系数矩阵
    
    
    correlation coefficient
    
    Parameter
    ---------
    *arg: list
    
    Attribute
    ---------
    rc_mat: correlation coefficient matrix
    '''
    def __init__(self, *arg):
        arg = np.array(arg, dtype=float)
        if len(arg.shape) != 2:
            raise Exception("Input list should be 1-D.")
        else:
            self.rc_mat = _rc(*arg)
    
    
    def __img_process(self, index, cmap):
        plt.matshow(self.rc_mat, cmap=cmap)
        plt.colorbar()
        if index:
            n_list = range(len(index))
            plt.xticks(n_list, index)
            plt.yticks(n_list, index)
    
    
    def show(self, index=None, cmap='Blues'):
        '''
        作图并显示
        
        参数
        ----
        index：列表形式，可选，各数组名称
        cmap：字符串形式，可选，颜色板，默认为'Blues'
        
        
        Display the image
        
        Parameters
        ----------
        index: list, callable, names of each array
        cmap: str, callable, color board, default='Blues'
        '''
        RC.__img_process(self, index=index, cmap=cmap)
        plt.show()
    
    
    def savefig(self, filename, index=None, cmap='Blues'):
        '''
        作图并保存
        
        参数
        ----
        filename：字符串形式，文件名
        index：列表形式，可选，各数组名称
        cmap：字符串形式，可选，颜色板，默认为'Blues'
        
        
        Save the image
        
        Parameters
        ----------
        filename: str, file name
        index: list, callable, names of each array
        cmap: str, callable, color board, default='Blues'
        '''
        RC.__img_process(self, index=index, cmap=cmap)
        plt.savefig(filename)


'''
本模块用于处理时间数据

This module is used for processing time data
'''
import numpy as np


def time_before(time_list, time, itself=False, sep=True):
    '''
    寻找特定时间
    在时间列表中寻找某个时间点之前的时间
    要求time参数中时间的表述为'xxxx/xx/xx'，分别对应年-月-日，/表示可以是任意符号
    如“1970年1月1日”应表述为'1970/01/01'或'1970.01.01'等
    
    参数
    ----
    time_list：列表类型，时间列表
    time：字符串类型，设定的时间点
    itself：布尔类型，寻找时是否包括设定的时间点本身
    sep：布尔类型，time_list中的元素是否有符号将年月日分开，默认为True
    
    返回
    ----
    元组形式，(筛选到的时间(列表), 筛选到的时间索引(列表))
    
    
    Search special time
    search all then time in time list which are before the point you set
    the expression of time are required to be like 'xxxx/xx/xx', corresponding to year-month-date, / can be any symbol
    e.g.  "January 1st, 1970" should be turned to '1970/01/01' or '1970.01.01' etc
    
    Parameters
    ----------
    time_list: list, list of time
    time: str, the time point
    itself: bool, if include the time point itself when searching
    sep: bool, True means elements in time_list have symbols to seperate the year, month and day, default=True
    
    Return
    ------
    tuple, (filtered time(list), index of the filtered time(list))
    '''
    select_time = []
    select_index = []
    date = int(time[:4]+time[5:7]+time[8:10])
    
    for i in range(len(time_list)):
        judge_time = time_list[i]
        if sep:
            judge_num = int(judge_time[:4] + judge_time[5:7] + judge_time[8:10])
        else:
            judge_num = int(judge_time)
        
        if date > judge_num:
            select_time.append(judge_time)
            select_index.append(i)
        
        if itself:
            if date == judge_num:
                select_time.append(judge_time)
                select_index.append(i)
    
    return select_time, select_index


def time_after(time_list, time, itself=False, sep=True):
    '''
    寻找特定时间
    在时间列表中寻找某个时间点之后的时间
    要求时间的表述为'xxxx/xx/xx'，分别对应年-月-日，/表示可以是任意符号
    如“1970年1月1日”应表述为'1970/01/01'或'1970.01.01'等
    
    参数
    ----
    time_list：列表类型，时间列表
    time：字符串类型，设定的时间点
    itself：布尔类型，寻找时是否包括设定的时间点本身
    sep：布尔类型，time_list中的元素是否有符号将年月日分开，默认为True
    
    返回
    ----
    元组形式，(筛选到的时间(列表), 筛选到的时间索引(列表))
    
    
    Search special time
    search all then time in time list which are after the point you set
    the expression of time are required to be like 'xxxx/xx/xx', corresponding to year-month-date, / can be any symbol
    e.g.  "January 1st, 1970" should be turned to '1970/01/01' or '1970.01.01' etc
    
    Parameters
    ----------
    time_list: list, list of time
    time: str, the time point
    itself: bool, if include the time point itself when searching
    sep: bool, True means elements in time_list have symbols to seperate the year, month and day, default=True
    
    Return
    ------
    tuple, (filtered time(list), index of the filtered time(list))
    '''
    select_time = []
    select_index = []
    date = int(time[:4] + time[5:7] + time[8:10])
    
    for i in range(len(time_list)):
        judge_time = time_list[i]
        if sep:
            judge_num = int(judge_time[:4] + judge_time[5:7] + judge_time[8:10])
        else:
            judge_num = int(judge_time)
        
        if judge_num > date:
            select_time.append(judge_time)
            select_index.append(i)
        
        if itself:
            if date == judge_num:
                select_time.append(judge_time)
                select_index.append(i)
    
    return select_time, select_index


def time_equal(time_list, time, sep=True, equal=True):
    '''
    寻找特定时间的索引
    在时间列表中寻找某个时间点
    要求时间的表述为'xxxx/xx/xx'，分别对应年-月-日，/表示可以是任意符号
    如“1970年1月1日”应表述为'1970/01/01'或'1970.01.01'等
    
    参数
    ----
    time_list：列表类型，时间列表
    time：列表类型，设定的时间点列表
    sep：布尔类型，time_list中的元素是否有符号将年月日分开，默认为True
    equal：布尔类型，True表示搜索与time相同的时间，False表示搜索与time不相同的时间
    
    返回
    ----
    元组形式，(筛选到的时间(列表), 筛选到的时间索引(列表))
    
    
    Search index of special time
    search all then time in time list which are the same as the point you set
    the expression of time are required to be like 'xxxx/xx/xx', corresponding to year-month-date, / can be any symbol
    e.g.  "January 1st, 1970" should be turned to '1970/01/01' or '1970.01.01' etc
    
    Parameters
    ----------
    time_list: list, list of time
    time: list, the time point list
    sep: bool, True means elements in time_list have symbols to seperate the year, month and day, default=True
    equal: bool, True means searching the time in "time", False means searching the time not in "time"
    
    Return
    ------
    tuple, (filtered time(list), index of the filtered time(list))
    '''
    select_index = []
    select_time = []
    for i in range(len(time)):
        time[i] = int(time[i][:4] + time[i][5:7] + time[i][8:10])
    
    if equal:
        for i in range(len(time_list)):
            judge_time = time_list[i]
            if sep:
                judge_num = int(judge_time[:4] + judge_time[5:7] + judge_time[8:10])
            else:
                judge_num = int(judge_time)
            
            if judge_num in time:
                select_index.append(i)
                select_time.append(judge_time)
    
    else:
        for i in range(len(time_list)):
            judge_time = time_list[i]
            if sep:
                judge_num = int(judge_time[:4] + judge_time[5:7] + judge_time[8:10])
            else:
                judge_num = int(judge_time)
            
            if not judge_num in time:
                select_index.append(i)
                select_time.append(judge_time)
        
        return select_time, select_index


def time_between(time_list, begin, end, begin_itself=False, end_itself=False, sep=True):
    '''
    寻找特定时间
    在时间列表中寻找某个时间点之后的时间
    要求时间的表述为'xxxx/xx/xx'，分别对应年-月-日，/表示可以是任意符号
    如“1970年1月1日”应表述为'1970/01/01'或'1970.01.01'等
    
    参数
    ----
    time_list：列表类型，时间列表
    begin：字符串类型，设定的开始时间点
    end：字符串类型，设定的结束时间点
    begin_itself：布尔类型，寻找时是否包括设定的开始时间点本身
    end_itself：布尔类型，寻找时是否包括设定的结束时间点本身
    sep：布尔类型，time_list中的元素是否有符号将年月日分开，默认为True
    
    返回
    ----
    元组形式，(筛选到的时间(列表), 筛选到的时间索引(列表))
    
    
    Search special time
    search all then time in time list which are after the point you set
    the expression of time are required to be like 'xxxx/xx/xx', corresponding to year-month-date, / can be any symbol
    e.g.  "January 1st, 1970" should be turned to '1970/01/01' or '1970.01.01' etc
    
    Parameters
    ----------
    time_list: list, list of time
    begin: str, the beginning time point
    end: str, the end time point
    bengin_itself: bool, if include the beginning time point itself when searching
    end_itself: bool, if include the end time point itself when searching
    sep: bool, True means elements in time_list have symbols to seperate the year, month and day, default=True
    
    Return
    ------
    tuple, (filtered time(list), index of the filtered time(list))
    '''
    select_time = []
    select_index = []
    
    # begin
    date_1 = int(begin[:4] + begin[5:7] + begin[8:10])
    
    # end
    date_2 = int(end[:4] + end[5:7] + end[8:10])
    
    if not begin_itself and not end_itself:
        for i in range(len(time_list)):
            judge_time = time_list[i]
            if sep:
                judge_num = int(judge_time[:4] + judge_time[5:7] + judge_time[8:10])
            else:
                judge_num = int(judge_time)
            
            if date_1 < judge_num < date_2:
                select_time.append(judge_time)
                select_index.append(i)
    
    elif begin_itself and not end_itself:
        for i in range(len(time_list)):
            judge_time = time_list[i]
            if sep:
                judge_num = int(judge_time[:4] + judge_time[5:7] + judge_time[8:10])
            else:
                judge_num = int(judge_time)
            
            if date_1 <= judge_num < date_2:
                select_time.append(judge_time)
                select_index.append(i)
    
    elif not begin_itself and end_itself:
        for i in range(len(time_list)):
            judge_time = time_list[i]
            if sep:
                judge_num = int(judge_time[:4] + judge_time[5:7] + judge_time[8:10])
            else:
                judge_num = int(judge_time)
            
            if date_1 < judge_num <= date_2:
                select_time.append(judge_time)
                select_index.append(i)
    
    else:
        for i in range(len(time_list)):
            judge_time = time_list[i]
            if sep:
                judge_num = int(judge_time[:4] + judge_time[5:7] + judge_time[8:10])
            else:
                judge_num = int(judge_time)
            
            if date_1 <= judge_num <= date_2:
                select_time.append(judge_time)
                select_index.append(i)
    
    return select_time, select_index


def _time_adjust(time):
    '''
    调整时间至正确格式
    '''
    for i in range(2):
        decimal = time[i] - int(time[i])
        time[i] -= decimal
        time[i+1] += decimal * 60
    
    for i in range(1, 3):
        delta = time[-i] // 60
        time[-i] = time[-i] % 60
        time[-i-1] += delta
    
    if time[0] < 0:
        for i in range(1, 3):
            if time[-i] > 0:
                time[-i] -= 60
                time[-i-1] += 1
    return time


def time_add(*time):
    '''
    时间相加
    例：
        time_add(time_1, time_2, ..., time_n)
        得到Σ time_i
    
    参数
    ----
    time：列表形式，时间，格式为[时, 分, 秒]
    
    返回
    ----
    一维ndarray，相加后的时间
    
    
    Time Addition
    e.g.
        time_add([1, 23, 45], [5, 43, 21])
        return array([7, 7, 6])
    
    Parameter
    ---------
    time: list, time, in the form of [hour, minute, second]
    
    Return
    ------
    1-D ndarray, time after addition
    '''
    time = np.array(time)
    time = time.sum(axis=0)
    return _time_adjust(time)


def time_sub(time_1, time_2):
    '''
    时间相减
    例：
        time_sub(time_1, time_2)
        得time_1 - time_2
    
    参数
    ----
    time_1，time_2：列表形式，时间，格式为[时, 分, 秒]
    
    返回
    ----
    一维ndarray，相减后的时间
    
    
    Time Subtraction
    e.g.
        time_sub(time_1, time_2)
        return time_1 - time_2
    
    Parameters
    ----------
    time_1，time_2: list, time, in the form of [hour, minute, second]
    
    Return
    ------
    1-D ndarray, time after subtraction
    '''
    time_1 = np.array(time_1)
    time_2 = -np.array(time_2)
    return time_add(time_1, time_2)


def time_mul(time, num):
    '''
    时间与数相乘
    例：
        time_sub(time, num)
        得time * num
    
    参数
    ----
    time：列表形式，时间，格式为[时, 分, 秒]
    num：数
    
    返回
    ----
    一维ndarray，相乘后的时间
    
    
    Time Multiplication
    e.g.
        time_sub(time, num)
        return time * num
    
    Parameters
    ----------
    time: list, time, in the form of [hour, minute, second]
    num: num
    
    Return
    ------
    1-D ndarray, time after multiplication
    '''
    time = np.array(time) * np.array(num)
    return _time_adjust(time)


def time_div(time, divisor, time_mode=False):
    '''
    时间相除
    时间除以数或时间除以时间
    
    参数
    ----
    time：列表形式，时间，格式为[时, 分, 秒]
    divisor：数或列表，数或时间，时间格式同上
    time_mode：布尔类型，可选，时间模式，True表示divisor要输入时间，False表示divisor要输入数，默认为False
    
    返回
    ----
    time_mode=False时返回数，time_mode=True时返回一维ndarray
    
    
    Time Division
    time divided by number or time divided by time
    
    Parameters
    ----------
    time: list, time, in the form of [hour, minute, second]
    divisor: num or list, number or time, in the form of [hour, minute, second]
    time_mode: bool, callable, time mode, True means divisor needs a time as input, while False needs a number, default=False 
    
    Return
    ------
    return a number when time_mode=False, return 1-D ndarray when time_mode=True
    '''
    time = np.array(time)
    if not time_mode:
        time = time / divisor
        return _time_adjust(time)
    else:
        divisor = np.array(divisor)
        divisor = 3600 * divisor[0] + 60 * divisor[1] + divisor[0]
        time = 3600 * time[0] + 60 * time[1] + time[0]
        return time / divisor


'''
该模块用于统计分析

This module is uesd for statistical analysis
'''
import numpy as np
from matplotlib import pyplot as plt
from scipy.integrate import quad
from scipy.stats import chi2
from scipy.optimize import minimize
from scipy.interpolate import interp1d


def r_moment(data, a=None, b=None, k=1):
    '''
    计算原点矩
    
    参数
    ----
    data：一维列表或函数类型，样本或概率密度函数
    a：浮点数类型，可选，计算下限
    b：浮点数类型，可选，计算上限
    k：整型，可选，阶数
    
    返回
    ----
    数，原点矩
    
    
    Calculate raw moment
    
    Parameters
    ----------
    fdata: 1-D list or function, a sample or probability density function
    a: float, callable, lower limit of calculation
    b: float, callable, upper limit of calculation
    k: int, callable, orders
    
    Return
    ------
    num, raw moment
    '''
    if type(data).__name__ == 'function':
        def ex_func(x):
            return x**k * data(x)
        
        return quad(ex_func, a, b)[0]
    
    else:
        data = np.array(data)**k
        return data.mean()


def ex(data, a=None, b=None):
    '''
    计算数学期望，等价于一阶原点矩
    
    参数
    ----
    data：一维列表或函数类型，样本或概率密度函数
    a：浮点数类型，可选，计算下限
    b：浮点数类型，可选，计算上限
    
    返回
    ----
    数，数学期望
    
    
    Calculate mathematical expectation, which is equivalent to one order raw moment
    
    Parameters
    ----------
    fdata: 1-D list or function, a sample or probability density function
    a: float, callable, lower limit of calculation
    b: float, callable, upper limit of calculation
    
    Return
    ------
    num, mathematical expectation
    '''
    if type(data).__name__ == 'function':
        def ex_func(x):
            return x * data(x)
        
        return quad(ex_func, a, b)[0]
    
    else:
        data = np.array(data)
        return data.mean()


def c_moment(data, k, a=None, b=None):
    '''
    计算中心矩
    
    参数
    ----
    data：一维列表或函数类型，样本或概率密度函数
    a：浮点数类型，可选，计算下限
    b：浮点数类型，可选，计算上限
    k：整型，可选，阶数
    
    返回
    ----
    数，中心矩
    
    
    Calculate central moment
    
    Parameters
    ----------
    data: 1-D list or function, a sample or probability density function
    a: float, callable, lower limit of calculation
    b: float, callable, upper limit of calculation
    k: int, callable, orders
    
    Return
    ------
    num, central moment
    '''
    raw_exp = ex(data, a, b)
    
    if type(data).__name__ == 'function':
        def ex_func(x):
            return (x-raw_exp)**k * data(x)
        
        return quad(ex_func, a, b)[0]
    
    else:
        data = np.array(data) - raw_exp
        return r_moment(data, k=k)


def dx(data, a=None, b=None):
    '''
    计算方差，等价于二阶中心矩
    
    参数
    ----
    data：一维列表或函数类型，样本或概率密度函数
    a：浮点数类型，可选，计算下限
    b：浮点数类型，可选，计算上限
    
    返回
    ----
    数，方差
    
    
    Calculate variance, which is equivalent to two orders central moment
    
    Parameters
    ----------
    data: 1-D list or function, a sample or probability density function
    a: float, callable, lower limit of calculation
    b: float, callable, upper limit of calculation
    
    Return
    ------
    num, variance
    '''
    return c_moment(data, 2, a, b)


def skew(data, a=None, b=None):
    '''
    计算偏度，等价于三阶中心矩
    
    参数
    ----
    data：一维列表或函数类型，样本或概率密度函数
    a：浮点数类型，可选，计算下限
    b：浮点数类型，可选，计算上限
    
    返回
    ----
    数，偏度
    
    
    Calculate skewness, which is equivalent to three orders central moment
    
    Parameters
    ----------
    data: 1-D list or function, a sample or probability density function
    a: float, callable, lower limit of calculation
    b: float, callable, upper limit of calculation
    
    Return
    ------
    num, skewness
    '''
    return c_moment(data, 3, a, b)


def kurt(data, a=None, b=None):
    '''
    计算峰度，等价于四阶中心矩
    
    参数
    ----
    data：一维列表或函数类型，样本或概率密度函数
    a：浮点数类型，可选，计算下限
    b：浮点数类型，可选，计算上限
    
    返回
    ----
    数，峰度
    
    
    Calculate kurtosis, which is equivalent to four orders central moment
    
    Parameters
    ----------
    data: 1-D list or function, a sample or probability density function
    a: float, callable, lower limit of calculation
    b: float, callable, upper limit of calculation
    
    Return
    ------
    num, kurtosis
    '''
    return c_moment(data, 4, a, b)


def mle(func, data, num, p_range=(-1, 1), method=None, tol=None):
    '''
    最大似然法
    求解似然函数L(θ) = ln(∏ f(xi; θ)) = ∑ ln(f(xi; θ))在θ ∈ p_range范围内的最大值

    参数
    ----
    func：函数类型，密度概率函数，func的形式应如下：
        def func(x, param):
        其中x需输入一个数或一个向量，是随机变量；param是一维ndarray，为未确定的参数
    data：一维或二维列表，样本
    num：未确定参数个数
    p_range：一维或二维列表，未确定参数的取值范围，若为一维列表，则代表所有未确定参数的取值范围都一样，默认为(-1, 1)
    method：字符串类型，求最大值的方法，可选'Nelder-Mead'、'Powell'、'CG'、'BFGS'、'Newton-CG'、'L-BFGS-B'、'TNC'、'COBYLA'、'SLSQP'、'trust-constr'、'dogleg'、'trust-ncg'、'trust-exact'、'trust-krylov'
    tol：误差
    
    返回
    ----
    1-D ndarray，最佳的参数值


    Maximum Likelihood Estimate
    Find the maximum of likelihood function L(θ) = ln(∏ f(xi; θ)) = ∑ ln(f(xi; θ)) when θ ∈ p_range

    Parameters
    ----------
    func：function, probability density function, the form of func should be as follows：
        def func(x, param):
        x is a number or a vector, random variable; param is 1-D ndarray, undetermined parameters
    data：1-D or 2-D list，samples
    num：the number of undetermined parameters
    p_range：1-D or 2-D list, value range of undetermined parameters, in case of one-dimensional list，all the undetermined parameters are in the same range, default=(-1, 1)
    method：str，the method of finding the maximum, 'Nelder-Mead'、'Powell'、'CG'、'BFGS'、'Newton-CG'、'L-BFGS-B'、'TNC'、'COBYLA'、'SLSQP'、'trust-constr'、'dogleg'、'trust-ncg'、'trust-exact'、'trust-krylov' are optional
    tol：error
    
    Return
    ------
    1-D ndarray, the best parameter value
    '''
    # 构建似然函数的相反数
    def Lmin(theta):
        result = 0
        for i in data:
            result += np.log(func(i, theta))
        return result * (-1)

    x0 = []
    constraints = []
    p_range = np.array(p_range)
    if len(p_range.shape) == 1:
        p_range = np.tile(p_range, (num, 1))
    for i in range(num):
        x0.append(p_range[i][0])
        constraints.append({'type':'ineq', 'fun':lambda x: x[i] - p_range[i][0]})
        constraints.append({'type':'ineq', 'fun':lambda x: p_range[i][1] - x[i]})
    x0 = np.array(x0)
        
    res = minimize(Lmin, x0, method=method, constraints=constraints, tol=tol)
        
    return res.x


class DF1d():
    '''
    一维分布拟合
    用频率分布函数逼近概率密度函数
    
    参数
    ----
    sample：1-D列表，样本点
    span：1-D列表，区间间隔，如span = [a, b, c]则将区间分为[a, b]和[b, c]，并统计各区间频率
    kind：浮点数类型或整型，可选，将插值类型指定为字符串('linear'、'nearest'、'zero'、'slinear'、'squardic'、'previous'、'next'，其中'zero'、'slinear'、'squared'和'cubic'表示零阶、一阶、二阶或三阶样条曲线插值；'previous'和'next'只返回点的上一个或下一个值)或作为一个整数指定要使用的样条曲线插值器的顺序。默认为'linear'
    
    属性
    ----
    f：概率密度函数
    
    
    One-dimension distribution fitting
    Approximation of probability density function by frequency distribution function
    
    Parameters
    ----------
    sample: 1-D list, samples
    span: 1-D list, span of intervals, e.g. span = [a, b, c] will divide the interval into [a, b] and [b, c], and count the frequency of each interval
    kind: float or int, callable, specifies the kind of interpolation as a string ('linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic', 'previous', 'next', where 'zero', 'slinear', 'quadratic' and 'cubic' refer to a spline interpolation of zeroth, first, second or third order; 'previous' and 'next' simply return the previous or next value of the point) or as an integer specifying the order of the spline interpolator to use. default='linear'
    
    Attribution
    -----------
    f: probability density function
    '''
    def __init__(self, sample, span, kind='linear'):
        x_list = []
        y_list = []
        sample = np.array(sample)
        total_num = len(sample)
        for i in range(len(span)-1):
            num = len(sample[sample < span[i+1]]) - len(sample[sample < span[i]])
            y_list.append(num / total_num)
            x_list.append(0.5 * (span[i] + span[i+1]))
        self.f = interp1d(x_list, y_list, kind=kind)
        self.__minmax = [min(x_list), max(x_list)]
    
    
    def show(self, acc=0.01):
        '''
        作图并显示
        
        参数
        ----
        acc：浮点数类型，精度
        
        
        Display the image
        
        Parameters
        ----------
        acc: float, accuracy
        '''
        x = np.arange(self.__minmax[0], self.__minmax[1], acc)
        plt.plot(x, self.f(x))
        plt.show()
    
    
    def savefig(self, filename, acc=0.01):
        '''
        作图并保存
        
        参数
        ----
        filename：字符串类型，保存的文件名
        acc：浮点数类型，精度
        
        
        Save the image
        
        Parameters
        ----------
        filename: str, file name
        acc: float, accuracy
        '''
        x = np.arange(self.__minmax[0], self.__minmax[1], acc)
        plt.plot(x, self.f(x))
        plt.savefig(filename)


class DFT():
    '''
    单个分布拟合检验
    
    参数
    ----
    func_type：字符串类型，分布拟合采用的函数的类型，可选'pdf'(概率密度函数)、'cdf'(分布函数)和'dis'(离散分布函数)，默认为'pdf'
    
    属性
    ----
    degree：卡方分布的自由度
    chi2_value：卡方值
    P：拒绝假设的错误概率
    
    
    Single distribution fitting test

    Parameter
    ---------
    func_type: str, callable, types of functions used in distribution fitting, 'pdf'(probability density function), 'cdf'(cumulative distribution function) and 'dis'(discrete distribution function) are optional, default='pdf'
    
    Atrributes
    ----------
    degree: degrees of freedom of chi square distribution
    chi2_value: chi square value
    P: Probability of error in rejecting hypothesis
    '''
    def __init__(self, func_type=None):
        if not func_type:
            self.__func_type = 'pdf'
        else:
            self.__func_type = func_type
    
    
    def __merge(self, data_list, pro_list):
        # 找到npi < 5的元素位置
        delete_index = np.where(pro_list < 5)[0]
        if len(delete_index) > 0:
            delete_pro = pro_list[delete_index]
            delete_data = data_list[delete_index]
            pro_final = [delete_pro[0]]
            data_final = [delete_data[0]]
            for i in range(1, len(delete_pro)):
                if pro_final[-1] < 5:
                    pro_final[-1] += delete_pro[i]
                    data_final[-1] += delete_data[i]
                else:
                    pro_final.append(delete_pro[i])
                    data_final.append(delete_data[i])
            pro_final = np.array(pro_final)
            data_final = np.array(data_final)
            
            # 删除原列表npi < 5位置对应的元素
            data_list = np.delete(data_list, delete_index)
            pro_list = np.delete(pro_list, delete_index)
            
            data_list = np.hstack((data_list, data_final))
            pro_list = np.hstack((pro_list, pro_final))
            
            if len(pro_list) == 1:
                raise Exception("Degree of freedom of chi-square distribution is 0.")
            
            elif len(pro_list) > 1 and pro_list[-1] < 5:
                data_list[-2] += data_list[-1]
                pro_list[-2] += pro_list[-1]
                data_list = np.delete(data_list, -1)
                pro_list = np.delete(pro_list, -1)
            
        # 计算fi^2数组
        data_list = data_list**2
        
        return data_list, pro_list
    
    
    def __con_fit(self, data, func, num_data):
        data = sorted(data)
        
        # 从小到大按20个一组分成多个部分
        data_list = []
        for i in range(int(num_data / 20 + 1)):
            data_list.append(data[i*20: (i+1)*20])
        
        # 如果最后一个部分的数量只有0个或1个元素，则叠加到前面
        if len(data_list[-1]) == 1:
            data_list[-2].append(data_list[-1][0])
            del data_list[-1]
        
        elif len(data_list[-1]) == 0:
            del data_list[-1]
        
        # 计算npi数组
        pro_list = []
        if self.__func_type == 'pdf':
            for i in data_list:
                pro_list.append(quad(func, i[0], i[-1])[0])
        
        elif self.__func_type == 'cdf':
            for i in data_list:
                pro_list.append(func(i[-1]) - func(i[0]))
        
        pro_list = np.array(pro_list) * num_data
        
        # 将数据数组转化为fi数组
        for i in range(len(data_list)):
            data_list[i] = len(data_list[i])
        data_list = np.array(data_list)
        
        data_list, pro_list = DFT.__merge(self, data_list, pro_list)

        return data_list, pro_list
    

    def __dis_fit(self, data, func, num_data):
        data = list(data)
        data_set = sorted(set(data))

        # 求出各data中个元素出现的次数并计算npi数组
        data_list = []
        pro_list = []
        for i in data_set:
            data_list.append(data.count(i))
            pro_list.append(func(i))
        data_list = np.array(data_list)
        pro_list = np.array(pro_list) * num_data

        data_list, pro_list = DFT.__merge(self, data_list, pro_list)

        return data_list, pro_list
    
    
    def fit(self, data, func):
        '''
        参数
        ----
        data：数据
        func：函数类型，分布函数或概率密度函数，函数的输入须为一个数
        
        
        Parameters
        ----------
        data: data
        func: function, cumulative distribution function or probability density function, the input of func should be a number
        '''
        self.__data = np.array(data, dtype=np.float)
        self.__func = func
        num_data = len(data)
        
        if self.__func_type != 'dis':
            data_list, pro_list = DFT.__con_fit(self, data, func, num_data)
        else:
            data_list, pro_list = DFT.__dis_fit(self, data, func, num_data)
        
        self.degree = len(pro_list) - 1
        self.chi2_value = sum(data_list / pro_list) - num_data
        self.P = 1 - chi2.cdf(self.chi2_value, self.degree)
    
    
    def __image_process(self, acc):
        x_min = min(self.__data)
        x_max = max(self.__data)
        x = np.arange(x_min, x_max, acc)
        y = self.__func(x)
        
        func_type = type(y).__name__
        if func_type == 'int' or func_type == 'float':
            y = np.ones_like(x) * 0.3
        else:
            y *= (0.3 / y.max())
        
        quartile1, medians, quartile3 = np.percentile(self.__data, [25, 50, 75])
        
        ax = plt.violinplot(self.__data, widths=0.6, showextrema=False, vert=False)
        ax['bodies'][0].set_facecolor('red')
        ax['bodies'][0].set_edgecolor('black')
        ax['bodies'][0].set_alpha(1)
        
        plt.scatter(medians, 1, marker='o', color='white', s=30, zorder=3)
        plt.hlines(1, quartile1, quartile3, color='k', linestyle='-', lw=5)
        plt.hlines(1, x_min, x_max, color='k', linestyle='-', lw=1)
        plt.vlines(self.__data.mean(), 0.9, 1.1, color='yellow', linestyle='-', lw=2)
        
        plt.plot(x, y)
        plt.yticks([])
    
    
    def show(self, acc=0.01):
        '''
        显示数据的小提琴图和假设的分布函数或概率密度图像
        
        参数
        ----
        acc：浮点数类型，可选，画图精度，默认为0.01
        
        
        Display the violin chart of data and the image of hypothetical cumulative distribution function or probability density function
        
        Parameter
        ---------
        acc: float, callable, the accuracy of drawing, default=0.01
        '''
        DFT.__image_process(self, acc)
        plt.show()
    
    def savefig(self,filename, acc=0.01):
        '''
        储存数据的小提琴图和假设的分布函数或概率密度图像
        
        参数
        ----
        filename：字符串类型，文件名
        acc：浮点数类型，可选，画图精度，默认为0.01
        
        
        Save the violin chart of data and the image of hypothetical cumulative distribution function or probability density function
        
        Parameters
        ----------
        filename: str, file name
        acc: float, callable, the accuracy of drawing, default=0.01
        '''
        DFT.__image_process(self, acc)
        plt.savefig(filename)


'''
本库用于科学计算和快速建模

maysics主要包括十二个模块：

1、algorithm 封装了几种模拟方法，用于简易模拟；
2、calculus 封装了部分常见的算符算子和积分方法，辅助数学运算；
3、constant 储存了部分常数；
4、equation 封装了部分方程求解运算；
5、explainer 用于评估和解释模型；
6、graph 用于图论分析；
7、models 封装了几种常用的数学物理定律、方程、模型以便快速构建数理模型；
8、preprocess 用于数据预处理；
9、stats 用于统计分析；
10、time_process 用于处理时间数据；
11、transformation 储存了常用的坐标转换及其他数学变换；
12、utils 是额外工具箱。


This package is used for scientific calculating and fast modeling.

maysics includes twelve modules:

1. "algorithm" packages several simulation methods for simple simulation;
2. "calculus" packages some common operators and integration method to assist in mathematical operations;
3. "constant" contents some usual constants;
4. "equation" packages some equation solving operation;
5. "explainer" used for estimating and explaining model;
6. "graph" used for graph theory analysis;
7. "models" packages several commonly used laws, equations and models of mathematical physics for fast modeling;
8. "preprocess" is used for data preproccessing;
9. "stats" is uesd for statistical analysis;
10. "time_process" is used for processing time data;
11. "transformation" stores common coordinate transformations and other mathematical transformations;
12. "utils" is extra Utils.
'''
import numpy as np
import pickle, csv
from PIL import Image
from maysics import algorithm, calculus, constant, equation, explainer, graph,\
    models, preprocess, stats, time_process, transformation, utils
from maysics.preprocess import preview, preview_file
from maysics.models import linear_r


def arr(f):
    '''
    将矢量函数的输出形式统一为ndarray
    
    参数
    ----
    f：函数类型
    
    返回
    ----
    更改输出格式后的函数
    
    
    transform the output of vector function as ndarray
    
    Parameter
    ---------
    f: function
    
    Return
    ------
    function after changing the output format
    '''
    def obj(x):
        func = f(x)
        try:
            return np.array(func)
        except:
            return func
    return obj


def add(*arg):
    '''
    实现函数与同型函数、函数与数之间的加法
    要求作用函数若输出列表，必须是ndarray格式
    
    返回
    ----
    相加后的新函数
    
    
    addition between function and function or function and number
    if output of the function is list, it requires ndarray
    
    Return
    ------
    new function after addition
    '''
    def obj(x):
        list = []
        for i in range(len(arg)):
            if type(arg[i]).__name__ == 'function':
                list.append(arg[i](x))
            else:
                list.append(arg[i])
        return sum(list)
    return obj


def mul(*arg):
    '''
    实现函数与同型函数、函数与数之间的乘法
    要求作用函数若输出列表，必须是ndarray格式
    
    返回
    ----
    相乘后的新函数
    
    
    multiplication between function and function or function and number
    if output of the function is list, it requires ndarray
    
    Return
    ------
    new function after multiplication
    '''
    def obj(x):
        result = 1
        for i in range(len(arg)):
            if type(arg[i]).__name__ == 'function':
                result *= arg[i](x)
            else:
                result *= arg[i]
        return result
    return obj


def sub(minuend, subtrahend):
    '''
    实现函数与同型函数、函数与数之间的减法
    要求作用函数若输出列表，必须是ndarray格式
    minuend：被减数
    subtrahend：减数
    
    返回
    ----
    相乘后的新函数
    
    
    subtraction between function and function or function and number
    if output of the function is list, it requires ndarray
    minuend: minuend
    subtrahend: subtrahend
    
    Return
    ------
    new function after subtraction
    '''
    def obj(x):
        if type(minuend).__name__ == 'function':
            result_of_minuend = minuend(x)
        else:
            result_of_minuend = minuend
        if type(subtrahend).__name__ == 'function':
            result_of_subtrahend = subtrahend(x)
        else:
            result_of_subtrahend = subtrahend
        
        result = result_of_minuend - result_of_subtrahend
        return result
    return obj


def div(dividend, divisor):
    '''
    实现函数与同型函数、函数与数之间的除法
    要求作用函数若输出列表，必须是ndarray格式
    
    参数
    ----
    dividend：被除数
    divisor：除数
    
    返回
    ----
    相乘后的新函数
    
    
    Parameters
    ----------
    division between function and function or function and number
    if output of the function is list, it requires ndarray
    dividend: dividend
    divisor: divisor
    
    Return
    ------
    new function after division
    '''
    def obj(x):
        if type(dividend).__name__ == 'function':
            result_of_dividend = dividend(x)
        else:
            result_of_dividend = dividend
        if type(divisor).__name__ == 'function':
            result_of_divisor = divisor(x)
        else:
            result_of_divisor = divisor
        result = result_of_dividend - result_of_divisor
        return result
    return obj


def covs1d(a, b, n):
    '''
    一维序列卷积和
    
    参数
    ----
    a：一维列表
    b：一维列表
    n：整型，平移步数
    
    返回
    ----
    数类型，a[n] * b[n]
    
    Convolution Sum of 1-D List
    
    Parameters
    ----------
    a: 1-D list
    b: 1-D list
    n: int, translation steps
    
    Return
    ------
    num, a[n] * b[n]
    '''
    a = np.array(a)
    b = list(b)
    b.reverse()
    b = np.array(b)
    num_a = len(a)
    num_b = len(b)
    if n <= 0 or n >= num_a + num_b:
        result = 0
    
    else:
        a = np.hstack((np.zeros(num_b), a, np.zeros(num_b)))
        b = np.hstack((b, np.zeros(num_a + num_b)))
        b[n : n+num_b] = b[: num_b]
        b[: n] = 0
        result = sum(a * b)
    
    return result


def covs2d(a, b, n, m):
    '''
    二维序列卷积和
    
    参数
    ----
    a：二维列表
    b：二维列表
    n：整型，沿axis=0方向的平移步数
    m：整型，沿axis=1方向的平移步数
    
    返回
    ----
    数类型，a[n, m] * b[n, m]
    
    Convolution Sum of 2-D List
    
    Parameters
    ----------
    a: 2-D list
    b: 2-D list
    n: int, translation steps along axis=0
    m: int, translation steps along axis=1
    
    Return
    ------
    num, a[n, m] * b[n, m]
    '''
    a = np.array(a)
    b = np.array(b)
    b = np.fliplr(b)
    b = np.flipud(b)
    num_a_x = a.shape[1]
    num_a_y = a.shape[0]
    num_b_x = b.shape[1]
    num_b_y = b.shape[0]
    
    if n <= 0 and m <= 0 or n >= num_a_x + num_b_x and m >= num_a_y + num_b_y:
        result = 0
    
    else:
        a = np.hstack((np.zeros((num_a_y, num_b_x)), a, np.zeros((num_a_y, num_b_x))))
        a = np.vstack((np.zeros((num_b_y, num_a_x + 2 * num_b_x)), a, np.zeros((num_b_y, num_a_x + 2 * num_b_x))))
        b = np.hstack((b, np.zeros((num_b_y, num_a_x + num_b_x))))
        b = np.vstack((b, np.zeros((num_a_y + num_b_y, num_a_x + 2 * num_b_x))))
        
        # 移动b矩阵
        b[n : n+num_b_y] = b[: num_b_y]
        b[: n] = 0
        b[:, m : m+num_b_x] = b[:, : num_b_x]
        b[:, : m] = 0
        
        result = (a * b).sum()
    
    return result


def save(filename, data, header=None):
    '''
    保存为.pkl、.npy或.csv文件
    
    参数
    ----
    filename：字符串类型，文件名
    data：需要保存的数据
    header：一维列表类型，可选，数据的列名称，仅在写入csv文件时有效
    
    
    Save as .pkl, .npy or .csv file
    
    Parameters
    ----------
    filename: str, file name
    data: data
    header: 1-D list, callable, the names of columns, effective only when writing csv files
    '''
    if filename[-4:] == '.pkl':
        with open(filename, 'wb') as file:
            pickle.dump(data, file)
    
    elif filename[-4:] == '.npy':
        np.save(filename, data)
    
    elif filename[-4:] == '.csv':
        data = np.array(data, dtype=np.object)
        if not header:
            header = []
            if len(data.shape) == 1:
                for i in range(data.shape[0]):
                    header.append(i)
                    with open(filename, 'w', newline='') as f:
                        writer = csv.writer(f)
                        writer.writerow(header)
                        writer.writerow(data)
            
            else:
                for i in range(data.shape[1]):
                    header.append(i)
                    with open(filename, 'w', newline='') as f:
                        writer = csv.writer(f)
                        writer.writerow(header)
                        writer.writerows(data)
    
    else:
        raise Exception("Suffix of filename must be '.pkl', '.npy' or '.csv'.")


def load(filename, header=True):
    '''
    载入.pkl、.npy或.csv文件
    
    参数
    ----
    filename：字符串类型，文件名
    header：布尔类型，可选，True表示csv文件第一行为列名，仅在读取csv文件时有效，默认为True
    
    
    Load .pkl, .npy or .csv file
    
    Parameter
    ---------
    filename: str, file name
    header: bool, callable, True means the first row of the csv file if the names of columns, effective only when reading csv files, default=True
    '''
    if filename[-4:] == '.pkl':
        with open(filename, 'rb') as file:
            data = pickle.load(file)
        
        return data
    
    elif filename[-4:] == '.npy':
        return np.load(filename, allow_pickle=True)
    
    elif filename[-4:] == '.csv':
        with open(filename, 'r') as f:
            reader = list(csv.reader(f))
            if header:
                reader = reader[1:]
            return np.array(reader)
    
    else:
        raise Exception("Suffix of filename must be '.pkl', '.npy' or '.csv'.")


def pic_data(filename, dtype=np.uint8):
    '''
    将图片转换为张量数据
    
    参数
    ----
    filename：字符串类型，文件名
    dtype：可选，返回的元素类型，默认为np.uint8
    
    返回
    ----
    ndarray，图片的张量数据
    
    
    Transform the picture into tensor
    
    Parameters
    ----------
    filename: str, file name
    dtype: callable, the type of elements, default=np.uint8
    
    Return
    ------
    ndarray, the tensor of the picture
    '''
    x = Image.open(filename)
    return np.array(x, dtype=dtype)


def data_pic(data, filename):
    '''
    将张量数据转换为图片并保存
    
    参数
    ----
    data：列表类型，张量数据，列表的元素要求是整型
    filename：字符串类型，文件名
    
    
    Transform the tensor to the picture and save the picture
    
    Parameters
    ----------
    data: list, tensor, the elements of the data are required to be int
    filename: str, file name
    '''
    data = np.array(data, dtype=np.uint8)
    image = Image.fromarray(data)
    image.save(filename)


def all_same(x):
    '''
    判断数组元素是否全相同
    
    参数
    ----
    x：数组
    
    返回
    ----
    布尔类型，True或者False
    
    
    Determine whether the array elements are all the same
    
    Parameter
    ---------
    x: array
    
    Return
    ------
    bool, True or False
    '''
    x = np.array(x)
    if len(x.shape) == 1:
        x = len(set(x))
        if x == 1:
            return True
        else:
            return False
    
    else:
        for i in x:
            if i.all() != x[0].all():
                return False
        return True


'''
本模块储存着一些常用的常数

This module contents some usual constants

数学 / math

chaos_1: 第一费根鲍姆常数 / the first Feigenbaum constant
chaos_2: 第二费根鲍姆常数 / the second Feigenbaum constant
e: 自然常数 / natural constant
gamma: 欧拉-马歇罗尼常数 / Euler-Mascheroni constant
golden: 黄金比例 / golden ratio
K: 兰道-拉马努金常数 / Landau–Ramanujan constant
K0: 辛钦常数 / Khintchine constant
pi: 圆周率 / Ratio of circumference to diameter
lp(): 勒让德多项式 / Legendre Polynomials
lpN()：勒让德多项式的模 / Module of Legendre Polynomials
alp()：连带勒让德多项式 / Associated Legendre Polynomials
alpN()：连带勒让德多项式的模 / Module of Associated Legendre Polynomials


物理 / physics

AU: 天文单位 / astronomical unit
atm: 标准大气压 / standard atmospheric pressure
c: 真空光速 / the speed of light in vacuum
c_e: 元电荷 / elementary charge
epsilon: 真空介电常数 / permittivity of vacuum
g: 重力加速度 / gravitational acceleration
G: 万有引力常数 / gravitational constant
h: 普朗克常数 / Planck constant
hr: 约化普朗克常数 / reduced Planck constant
k: 玻尔兹曼常数 / Boltzmann constant
lambdac: 康普顿波长 / Compton wavelength
ly: 光年 / light-year
m_e: 电子质量 / mass of electron
m_earth: 地球质量 / mass of the earth
m_n: 中子质量 / mass of neutron
m_p: 质子质量 / mass of proton
m_s: 太阳质量 / mass of the sun
miu: 真空磁导率 / permeability of vacuum
NA: 阿伏伽德罗常数 / Avogadro constant
pc: 秒差距 / Parsec
Platonic_year: 柏拉图年 / Platonic year
R: 理想气体常数 / gas constant
r_earth: 地球平均半径 / mean radius of the earth
r_sun: 太阳平均半径 / mean radius of the sun
r_e_m: 地月平均距离 / mean distance between the earth and the moon
SB: 斯特藩-玻尔兹曼常量 / Stefan-Boltzmann constant
v1: 第一宇宙速度 / first cosmic velocity
v2: 第二宇宙速度 / second cosmic velocity
v3: 第三宇宙速度 / third cosmic velocity
hp()：厄米多项式 / Hermite Polynomials
'''
import numpy as np
from scipy.special import factorial


# math
pi = 3.141592653589793
e = 2.718281828459045
golden = 1.618033988749895
gamma = 0.57721566490153286060651209
K = 0.76422365358922066299069873125
chaos_1 = 4.669201609102990
chaos_2 = 2.502907875095892
K0 = 2.6854520010


def lpN(l):
    '''
    l阶勒让德多项式的模
    
    参数
    ----
    l：整型，阶数
    
    返回
    ----
    数类型
    
    
    Module of Legendre Polynomials at degree l
    
    Paramater
    ---------
    l: int, degree
    
    Return
    ------
    num
    '''
    return (2 / (2 * l + 1))**0.5
    

def lp(x, l):
    '''
    l阶勒让德多项式的值
    
    参数
    ----
    x：数，输入值
    l：整型，阶数
    
    返回
    ----
    数
    
    
    Value of Legendre Polynomials at degree l
    
    Parameters
    ----------
    x: num, input value
    l: int, degree
    
    Return
    ------
    num
    '''
    result = 0
    for k in range(int(l/2) + 1):
        result += (-1)**k * factorial(2 * l - 2 * k)\
        / 2**l / factorial(k) / factorial(l - k) / factorial(l - 2 * k)\
        * x**(l - 2 * k)
    return result


def alpN(l, m):
    '''
    m阶l次连带勒让德多项式的模
    
    参数
    ----
    l：整型，次数
    m：整型，阶数
    
    Value of Associated Legendre Polynomials of order m and degree l
    
    Parameters
    ----------
    l: int, degree
    m: int, order
    '''
    up_part = 2 * factorial(l + m)
    down_part = (2 * l + 1) * factorial(l - m)
    return (up_part / down_part)**0.5


def _dm_lp(x, l, m, acc):
    '''
    勒让德多项式的m阶导数
    '''
    if m == 0:
        x = lp(x, l)
    elif m == 1:
        x = (lp(x + acc * 0.5, l) - lp(x - acc * 0.5, l)) / acc
    else:
        x = (_dm_lp(x + acc * 0.5, l, m-1, acc) - _dm_lp(x - acc * 0.5, l, m-1, acc)) / acc
    return x


def alp(x, l, m, acc=0.1):
    '''
    m阶l次连带勒让德多项式的值
    Plm(x) = (1-x^2)^(m/2) * d^m/dx^m * Pl(x)
    
    参数
    ----
    x：数，输入值
    l：整型，次数
    m：整型，阶数
    acc：浮点数类型，求导精度
    
    
    Value of Associated Legendre Polynomials of order m and degree l
    Plm(x) = (1-x^2)^(m/2) * d^m/dx^m * Pl(x)
    
    Parameters
    ----------
    x: num, input value
    l: int, degree
    m: int, order
    acc: float, callable, accuracy of derivation, default=0.1
    '''
    return (1 - x**2)**(m / 2) * _dm_lp(x, l, m, acc)


# physics
atm = 101325
c = 299792458.0
c_e = 1.602176634e-19
m_e = 9.1093837015e-31
m_earth = 5.965e24
m_n = 1.67492749804e-27
m_p = 1.67262192369e-27
m_sun = 1.9891e30
G = 6.6743e-11
g = 9.80665
h = 6.62607015e-34
hr = 1.0545718176461565e-34
k = 1.380649e-23
miu = 1.2566370614359173e-06
epsilon = 8.85e-12
NA = 6.02214179e23
R = 8.31446114959365
r_earth = 6371393
r_sun = 6.955e8
r_moon = 384403000
lambdac = 2.4263102175e-12
ly = 9460730472580800
AU = 149597870700
pc = 3.0856775814671915808e16
Platonic_year = 26000
SB = 5.670367e-8
v1 = 7900
v2 = 11200
v3 = 16700


def hp(x, n, li=False):
    '''
    n阶厄米多项式

    参数
    ----
    x：数，输入值
    n：整型，阶数
    li：布尔类型，可选，True表示将0到n阶勒让德多项式组成一维ndarray返回，默认为False

    返回
    ----
    数或一维ndarray


    Value of Hermite Polynomials at degree n

    Parameters
    ----------
    x: num, input value
    n: int, degree
    li: bool, callable, True means that 0 to n order Legendre polynomials are formed into 1-D ndarray and returned, default=False
    
    Reuturn
    -------
    num or 1-D ndarray
    '''
    if n <= 0:
        n = 0
    
    if not li:
        if n == 0:
            return 1

        elif n == 1:
            return 2 * x

        else:
            return 2 * x * hp(x, n-1) - 2 * (n-1) * hp(x, n-2)
    
    else:
        hp_list = []
        for i in range(n+1):
            if i == 0:
                hp_list.append(1)
            
            elif i == 1:
                hp_list.append(2 * x)
            
            else:
                hp_list.append(2 * x * hp_list[i-1] - 2 * (i-1) * hp_list[i-2])
        return np.array(hp_list)


'''
本模块储存着部分常用的数理物理方程、定律、数学模型等

This module stores some commonly used mathematical physical equations, laws, mathematical models, etc
'''
import numpy as np
from maysics import constant
from scipy.integrate import solve_ivp
from scipy.sparse import csr_matrix


def fouriers_law(T0, T, k, acc=0.1):
    '''
    傅里叶定律/热传导定律
    热流量：J=-k▽T
    
    参数
    ----
    T0：一维列表，某一位置的温度
    T：函数类型，温度分布函数
    k：浮点数类型，热导率
    acc：浮点数类型，可选，求导精度，默认为0.1
    
    返回
    ----
    数，热流量
    
    
    Fourier's Law: J=-k▽T
    
    Parameter
    ---------
    T0: 1-D list, the temperature of one point
    T: function, temperature distribution function
    k: float, coefficient of thermal conductivity
    acc: float, callable, accuracy of derivation, default=0.1
    
    Return
    ------
    num, heat flow
    '''
    T0 = np.array(T0, dtype=float)
    dim = len(T0)
    result = []
    for i in range(dim):
        T0[i] += acc * 0.5
        func1 = T(T0)
        T0[i] -= acc
        func2 = T(T0)
        T0[i] += acc * 0.5
        de = (func1 - func2) / acc
        result.append(de)
    result = np.array(result)
    return -k * result


def logistic(t, N0, r, K):
    '''
    Logistic人口增长模型
    该模型可得到解析解
    解的形式为：Nt = K*N0/(N0+(K-N0)*np.e**(-r*t))
    其中，Nt是t时刻的人口数
    
    参数
    ----
    t：时间
    N0：数，现有人口数
    r：数，人口自然增长率
    K：数，环境资源允许的稳定人口数
    
    返回
    ----
    数，Nt
    
    
    Logisyic population growth models
    solution: Nt = K*N0/(N0+(K-N0)*np.e**(-r*t))
    Nt is the population at time 't'
    
    Parameters
    ----------
    t: time
    N0: number or list, initial population
    r: number, natural population growth rate
    K: number, stable population allowed by environmental resources
    
    Return
    ------
    num, Nt
    '''
    Nt_pre_down_ = N0 + (K - N0) * np.e**(-r * t)
    Nt = K * N0 / Nt_pre_down_
    return Nt


def linear_r(x, y):
    '''
    线性回归模型
    
    参数
    ----
    x：一维或二维ndarray，自变量
    y：一维或二维ndarray，因变量
    
    返回
    ----
    一维ndarray：系数行矩阵
    损失函数：平均平方误差mse
    
    
    Linear Regression Model
    
    Parameters
    ----------
    x: 1-D or 2-D ndarray, independent variable
    y: 1-D or 2-D ndarray, dependent variable
    
    Return
    ------
    1-D ndarray: coefficient row matrix
    loss: mean squared error: mse
    '''
    x = np.array(x)
    if len(x.shape) == 1:
        x = np.array([x]).T
    y = np.array(y)
    x = np.hstack((x, np.ones((x.shape[0], 1))))
    coef = np.linalg.lstsq(x, y, rcond=None)[0].reshape(-1)
    y_pre = np.dot(x, np.array([coef]).T).reshape(y.shape)
    mse = ((y_pre - y)**2).mean()
    return coef, mse


class MVD_law():
    '''
    麦克斯韦速率分布律
    参数
    ----
    m：气体分子质量, 单位：kg
    T：气体温度, 单位：K
    
    属性
    ----
    fv：速率分布函数
    v_mean：平均速率
    v_p：最概然速率
    v_rms：均方根速率
    
    
    Maxwell's Velocity Distribution Law
    
    Parameters
    ----------
    m: mass of gas molecule, unit: kg
    T: temperature of gas, unit: K
    
    Attributes
    ----------
    fv: velocity distribution function
    v_mean: average velocity
    v_p: most probable velocity
    v_rms: root-mean-square velocity
    '''
    def __init__(self, m, T):
        def func_of_v(v):
            f_v_1 = 4 * np.pi * v**2
            f_v_2 = m / (2 * np.pi * constant.k * T)**1.5
            f_v_3 = np.e**(-m * v**2 / (2 * constant.k * T))
            return f_v_1 * f_v_2 * f_v_3
        
        self.fv = func_of_v
        v_part = (constant.k * T / (np.pi * m))**0.5
        self.v_mean = 8**0.5 * v_part
        self.v_p = 2**0.5 * v_part
        self.v_rms = 3**0.5 * v_part


class Plancks_law():
    '''
    普朗克黑体辐射定律
    
    参数
    ----
    T：黑体温度，单位：K
    
    属性
    ----
    Mf：频率形式的普朗克公式，频率单位：10^10 kHz
    Ml：波长形式的普朗克公式，波长单位：100nm
    
    
    Planck's Blackbody Radiation Law
    
    Parameter
    ---------
    T: temperature of blackbody, unit: K
    
    Attributes
    ----------
    Mf: Planck's formula in the form of frequency, unit of frequency: 10^10 kHz
    Ml: Planck's formula in the form of wave length, unit of wave length: 100 nm
    '''
    def __init__(self, T):
        h_k = constant.h / constant.k
        def Mf(f):
            f = f * 1e13
            f_1 = 2 * np.pi * constant.h * f**3 / constant.c**2
            f_2 = 1 / (np.e**(h_k * f / T) - 1)
            return f_1 * f_2
        
        def Ml(l):
            l = l * 1e-7
            l_1 = 2 * np.pi * constant.h * constant.c**2 / l**5
            l_2 = 1 / (np.e**(h_k * constant.c / (l * T)) - 1)
            return l_1 * l_2
        
        self.Mf = Mf
        self.Ml = Ml


class ED():
    '''
    传染病模型
    本类可构建SI模型、SIR模型、SIRS模型、SEIR模型
    S是易感染者
    E是暴露者，染病但还不具有传染能力
    I是感染者，带有传染能力
    R是康复者(带有免疫力)或死者
    
    参数
    ----
    I0：初始感染者
    K：样本总数
    beta：感染率
    
    
    mathematical models of epidemic diseases
    this class is used to buil SI, SIR, SIRS, SEIR
    S means susceptible people
    E means exposed people, which are infected without ability to infect
    I means infected people with ability to infect
    R means recovered people with immunity or dead people
    
    Parameters
    ----------
    I0: initial infected people
    K: sample size
    beta: infection rate
    '''
    def __init__(self, I0, K, beta, R=0, E=0):
        self.I0 = I0
        self.K = K
        self.beta = beta
        self.R = R
        self.E = E
        self.S = K - I0 - R - E
    
    
    def SI(self):
        '''
        SI模型：
        该模型不需要再额外输入参数，且可得到解析解
        解的形式为：I = K*(1-(K-I0)/(K+I0+I0*np.e**(beta*K*t)))

        返回
        ----
        函数形式，函数的返回值为(I(t), S(t))
        
        
        SI:
        no more parameters
        solution: I = K*(1-(K-I0)/(K+I0+I0*np.e**(beta*K*t)))
        
        Return
        ------
        function, the value returned by the function is (I(t), S(t))
        '''
        def obj(t):
            I_pre_down_ = self.K + self.I0 * (np.e**(self.beta * self.K * t) - 1)
            I_pre_up_ = self.K - self.I0
            I_func_ = self.K * (1 - I_pre_up_ / I_pre_down_)
            S_func_ = self.K - I_func_
            return I_func_, S_func_
        return obj
    
    
    def SIR(self, gamma, t_span, method='RK45' ,t_eval=None):
        '''
        SIR模型：
        该模型不能得到解析解，给出的是基于solve_ivp得到的数值解数组
        
        参数
        ----
        gamma：康复率
        t_span：元组形式，求解的上下限
        method：字符串形式，可选，求解方法，可选择'RK45'、'RK23'、'DOP835'、'Radau'、'BDF'、'LSODA'
        t_eval：数组形式，可选，每当t等于该数组中的值时，会生成一个数值解

        返回
        ----
        solve_ivp数值解，顺序是I, S, R
        
        
        SIR:
        
        Parameters
        ----------
        gamma: recovery rate
        t_span: tuple, limits of solution
        method: str, callable, solving method, 'RK45', 'RK23', 'DOP835', 'Radau', 'BDF', 'LSODA' are optional
        t_eval: list, callable, when 't' equals the value in the list, it will generate a numerical solution
        
        Return
        ------
        numerical solution of solve_ivp, the order is I, S, R
        '''
        y0 = [self.I0, self.S, self.R]
        def epis_equas(t, x):
            y1 = self.beta * x[0] * x[1] - gamma * x[0]
            y2 = -self.beta * x[0] * x[1]
            y3 = gamma * x[0]
            return y1, y2, y3
        result = solve_ivp(epis_equas, t_span=t_span, y0=y0, method=method, t_eval=t_eval)
        return result
    
    
    def SIRS(self, gamma, alpha, t_span, method='RK45' ,t_eval=None):
        '''
        SIRS模型：
        该模型不能得到解析解，给出的是基于solve_ivp得到的数值解数组
        
        参数
        ----
        gamma：康复率
        alpha：衡量康复者获得免疫的时间
        t_span：元组形式，求解的上下限
        method：字符串形式，可选，求解方法，可选择'RK45'、'RK23'、'DOP835'、'Radau'、'BDF'、'LSODA'
        t_eval：数组形式，可选，每当t等于该数组中的值时，会生成一个数值解

        返回
        ----
        solve_ivp数值解，顺序是I, S, R
        
        
        SIRS:
        
        Parameters
        ----------
        gamma: recovery rate
        alpha: measuring the time that the recovered people with immunity
        t_span: tuple, limits of solution
        method: str, callable, solving method, 'RK45', 'RK23', 'DOP835', 'Radau', 'BDF', 'LSODA' are optional
        t_eval: list, callable, when 't' equals the value in the list, it will generate a numerical solution
        
        Return
        ------
        numerical solution of solve_ivp, the order is I, S, R
        '''
        y0= [self.I0, self.S, self.R]
        def epis_equas(t, x):
            y1 = self.beta * x[0] * x[1] - gamma * x[0]
            y2 = -self.beta * x[0] * x[1] + alpha * x[2]
            y3 = gamma * x[0] - alpha * x[2]
            return y1, y2, y3
        result = solve_ivp(epis_equas, t_span=t_span, y0=y0, method=method, t_eval=t_eval)
        return result
    
    
    def SEIR(self, gamma1, gamma2, alpha, t_span, method='RK45' ,t_eval=None):
        '''
        SEIR模型：
        该模型不能得到解析解，给出的是基于solve_ivp得到的数值解数组
        
        参数
        ----
        gamma1：潜伏期康复率
        gamma2：患者康复率
        alpha：衡量康复者获得免疫的时间
        t_span：元组形式，求解的上下限
        method：字符串形式，可选，求解方法，可选择'RK45'、'RK23'、'DOP835'、'Radau'、'BDF'、'LSODA'
        t_eval：数组形式，可选，每当t等于该数组中的值时，会生成一个数值解

        返回
        ----
        solve_ivp数值解，顺序是I, S, R, E
        
        
        SEIR:
        
        Parameters
        ----------
        gamma1: recovery rate of incubation
        gamma2: recovery rate of patients
        alpha: measuring the time that the recovered people with immunity
        t_span: tuple, limits of solution
        method: str, callable, solving method, 'RK45', 'RK23', 'DOP835', 'Radau', 'BDF', 'LSODA' are optional
        t_eval: list, callable, when 't' equals the value in the list, it will generate a numerical solution
        
        Return
        ------
        numerical solution of solve_ivp, the order is I, S, R, E
        '''
        y0 = [self.I0, self.S, self.R, self.E]
        def epis_equas(t, x):
            y1 = alpha * x[3] - gamma2 * x[0]
            y2 = -self.beta * x[0] *x[1]
            y3 = gamma1 * x[3] + gamma2 * x[0]
            y4 = self.beta * x[0] * x[1] - (alpha + gamma1) * x[3]
            return y1, y2, y3, y4
        result = solve_ivp(epis_equas, t_span=t_span, y0=y0, method=method, t_eval=t_eval)
        return result


class Leslie():
    '''
    Leslie模型
    解的形式为：Nt = M**t * N0
    其中，Nt是t时刻的个体数列表，M是Leslie矩阵
    
    参数
    ----
    N0：列表，各年龄层初始个体数目
    r：数，各年龄层的生殖率
    s：数，各年龄层到下一个年龄层的存活率
    age_range：整型，年龄段的跨度，默认为1
    
    属性
    ----
    Leslie_matrix：莱斯利矩阵
    
    
    Leslie model
    solution: Nt = M**t * N0
    Nt is the list of individuals at time 't' , M is Leslie matrix
    
    Parameters
    ----------
    N0: list, initial number of individuals in each age group
    r: number, reproductive rate of each age group
    s: number, survival rate to next age group of each group
    age_range: int, the span of age groups, default=1
    
    Attribute
    ---------
    Leslie_matrix: Leslie matrix 
    '''
    def __init__(self, N0, r, s, age_range=1):
        self.N0 = np.mat(N0).T
        self.r = np.array(r)
        self.s = np.array(s)
        self.age_range = age_range
        
        Leslie_matrix = np.zeros((len(s), len(N0)))
        Leslie_matrix = np.vstack((r, Leslie_matrix))
        for i in range(1, len(N0)):
            Leslie_matrix[i][i-1] = s[i-1]
        self.Leslie_matrix = csr_matrix(Leslie_matrix)
    
    
    def predict(self, t):
        '''
        参数
        ----
        t：时间
        
        返回
        ----
        数，Nt
        
        
        Parameter
        ---------
        t: time
        
        Return
        ------
        num, Nt
        '''
        t_times = t // self.age_range
        dense_Leslie_matrix = self.Leslie_matrix.todense()
        Nt = dense_Leslie_matrix**t_times * self.N0
        return Nt


'''
本模块储存了蒙特卡洛模拟、遗传算法、模拟退火算法和梯度下降算法，用于简易模拟

This module stores Monte Carlo, Genetic Algorithm, Simulated Annealing Algorithm and Gradient Descent Algorithm for simple simulation
'''
import random
import numpy as np
from multiprocessing import Pool, cpu_count
from maysics.calculus import grad


class MC():
    '''
    蒙特卡洛模拟
    
    参数
    ----
    loop：整型，可选，循环次数，默认为1000
    random_type：字符串类型或函数类型，可选，可以取'random'、'randint'或自定义函数(无参数，需输出length * dim的二维ndarray)，默认'random'
    random_state：整型，可选，随机种子
    begin：整型，可选，表示整数随机序列的开始数字，仅当random_type='randint'时起作用
    end：整型，可选，表示整数随机序列的结束数字，仅当random_type='randint'时起作用
    n_jobs：整型，可选，调用的cpu数，-1表示全部调用，默认为1
    
    属性
    ----
    loop：整型，循环次数
    random_type：字符串类型，取随机的方法
    EX：数类型，select函数返回值的数学期望
    DX：数类型，select函数返回值的方差
    history：字典类型，历史EX和DX值
    
    
    Monte Carlo
    
    Parameters
    ----------
    loop: int, callable, loop count, default = 10000
    random_type: str or function, callable, 'random', 'randint' and function(no parameter, output a 2-D ndarray with size length * dim) are optional, default = 'random'
    random_state: int, callable, random seed
    begin: int, callable, beginning number of the random sequence, it's used only when random_type='randint'
    end: int, callable, end number of the random sequence, it's used only when random_type='randint'
    n_jobs: int, callable, the number of cpus called, -1 means to call all cpus

    Atrributes
    ----------
    loop: int, loop count
    random_type: str, the method of generating random number
    EX: num, mathematical expectation of return value of select function
    DX: num, variance of return value of select function
    history: dict, historical EX and DX
    '''
    def __init__(self, loop=1000, random_type='random', random_state=None,
                 begin=None, end=None, n_jobs=1):
        self.loop = loop
        self.random_type = random_type
        self.history = {'EX':[], 'DX':[]}
        self.__begin = begin
        self.__end = end
        self.n_jobs = n_jobs
        np.random.seed(random_state)
    

    def sel_ind(self, condition):
        '''
        用于快速构建条件函数
        将condition函数作用于随机序列全部元素
        任意一个状态满足condition，则会输出1，否则输出0
        
        参数
        ----
        condition：函数，判断一个状态是否符合条件的函数，要求符合则输出True，不符合则输出False

        返回
        ----
        一个新函数


        for quick construction of condition function
        apply condition function to all elements of random sequence
        if any state is qualified, 1 will be reutrn, otherwise 0 will be return
        
        Parameters
        ----------
        condition: function, a function that determines if a state is qualified. If qualified, return True, otherwise reutrn False

        Return
        ------
        a new function
        '''
        def obj(x):
            for i in x:
                if condition(i):
                    return 1
            return 0
        return obj

    
    def sel_any(self, n, condition):
        '''
        用于快速构建条件函数
        将condition函数作用于随机序列全部元素
        至少任意n个状态满足condition，则会输出1，否则输出0
        
        参数
        ----
        n：int类型
        condition：函数，判断一个状态是否符合条件的函数，要求符合则输出True，不符合则输出False

        返回
        ----
        一个新函数


        for quick construction of condition function
        apply condition function to all elements of random sequence
        if at least any_n states are qualified, 1 will be reutrn, otherwise 0 will be return
        
        Parameters
        ----------
        n: int
        condition: function, a function that determines if a state is qualified. If qualified, return True, otherwise reutrn False

        Return
        ------
        a new function
        '''
        def obj(x):
            calcu_ = 0
            for i in x:
                if condition(i):
                    calcu_ += 1
                if calcu_ >= n:
                    break
            if calcu_ >= n:
                return 1
            else:
                return 0
        return obj
    
    
    def sel_con(self, n, condition):
        '''
        用于快速构建条件函数
        将condition函数作用于随机序列全部元素
        至少连续n个状态满足condition，则会输出1，否则输出0
        
        参数
        ----
        n：int类型
        condition：函数，判断一个状态是否符合条件的函数，要求符合则输出True，不符合则输出False

        返回
        ----
        一个新函数


        for quick construction of condition function
        apply condition function to all elements of random sequence
        if at least con_n consecutive states are qualified, 1 will be reutrn, otherwise 0 will be return
        
        Parameters
        ----------
        n: int
        condtition: function, a function that determines if a state is qualified. If qualified, return True, otherwise reutrn False
        
        Return
        ------
        a new function
        '''
        def obj(x):
            calcu_ = 0
            for i in x:
                if condition(i):
                    calcu_ += 1
                else:
                    calcu_ *= 0
                if calcu_ >= n:
                    break
            if calcu_ >= n:
                return 1
            else:
                return 0
        return obj
    
    
    def __one_experiment(self, length, dim, select):
        '''
        进行一次完整的实验
        '''
        if self.random_type == 'random':
            random_list_ = np.random.rand(length, dim)

        elif self.random_type == 'randint':
            random_list_ = np.random.randint(self.__begin, self.__end + 1, size=(length, dim))
        
        else:
            random_list_ = self.random_type()

        #如果随机序列满足select，就算实验成功输出1，否则算失败输出0
        if type(select).__name__ == 'function':
            return select(random_list_)
        else:
            judge_ = 1
            for li in select:
                if li(random_list_):
                    judge_ *= 1
                else:
                    judge_ *= 0
            return judge_


    def fit(self, length, dim, select):
        '''
        进行蒙特卡洛模拟
        
        参数
        ----
        length：整型，随机序列长度
        dim：整型，随机序列元素维度
        select：条件函数或函数列表，select函数应该以length * dim的二维ndarray为自变量
            select函数用于甄别某一个随机矩阵是否符合预期，符合输出1，不符合输出0
            select函数同时也可以输出其他值以满足实际需求
            select若为函数列表，则要求每个函数只能输出1或0

        
        Simulate
        
        Parameters
        ----------
        length: int, the length of random sequence
        dim: int, dimension of elements in random sequence
        select: function or function list, select function should take 2-dimension ndarray(length * dim) as independent variable
            select function used for Identifying whether the random matrix meets the expectation, if meets, output 1, otherwise output 0
            select function can also output other values to meet the actual demand
            if select is a list, every function in the list can only output 1 or 0
        '''
        final_propability_ = []
        if self.n_jobs == 1:
            for i in range(self.loop):
                freq_ = MC.__one_experiment(self, length, dim, select)
                final_propability_.append(freq_)
        
        else:
            try:
                if self.n_jobs == -1:
                    pool = Pool(processes=cpu_count())
                else:
                    pool = Pool(processes=self.n_jobs)
                for i in range(self.loop):
                    freq_ = pool.apply(MC.__one_experiment,
                                     (self, length,dim, select))
                    final_propability_.append(freq_)
            
            except:
                for i in range(self.loop):
                    freq_ = MC.__one_experiment(self, length, dim, select)
                    final_propability_.append(freq_)
        
        self.EX = np.mean(final_propability_)
        self.DX = np.var(final_propability_)
        self.history['EX'].append(self.EX)
        self.history['DX'].append(self.DX)
    
    
    def clr(self):
        '''
        清空历史数据
        
        
        clear the historical data
        '''
        self.history = {'EX':[], 'DX':[]}


class GA():
    '''
    遗传算法
    
    参数
    ----
    population：整型，种群数
    iteration: 整型，迭代次数（自然选择次数）
    random_type：字符串类型，可选，可以取'random'和'randint'，默认'random'
    select：字符串类型或函数类型，可选，选择个体的方法，可选'rw'、'st'或者自定义函数，默认'rw'
        'rw'：基于随机接受的轮盘赌选择
        'st'：随机竞争选择
        自定义函数：函数需要有两个参数，第一个参数是一个二维ndarray，第二个参数是适应度函数
    crossover：字符串类型或函数类型，可选，交叉互换的方法，可选'uniform'、'point'或自定义函数，默认'uniform'
        'uniform'：均匀交叉
        'point'：单点及多点交叉
        自定义函数：函数只能设置一个参数，以种群（二维ndarray）作为输入
    begin：整型，可选，表示整数随机序列的开始数字，仅当random_type='randint'时起作用
    end：整型，可选，表示整数随机序列的结束数字，仅当random_type='randint'时起作用
    random_state：整型，可选，随机种子
    select_rate：浮点数类型，可选，选择率（存活率），默认0.3
    mutate_rate：浮点数类型，可选，变异率，默认0.05
    crossover_rate：浮点数类型，可选，基因交叉概率，默认0.5
    repeat：布尔类型，可选，是否允许序列元素重复，默认为True，仅在random_type='randint'且crossover不是自定义函数时起作用
    
    属性
    ----
    population：整型，种群数
    iteration：整型，迭代次数（自然选择次数）
    random_type：字符串类型，取随机的方法
    select：字符串类型，选择个体的方法
    crossover：字符串类型，交叉互换的方法
    dom：二维ndarray，优势种群
    dom_fitness：一维ndarray，优势种群的适应度
    
    
    Genetic Algorithm
    
    Parameters
    ----------
    population: int, the population size
    iteration: int, the times of iterations( the times of natural selection)
    random_type: str, callable, 'random' and 'randint' are optional, default = 'random'
    select: string or function, callable, the method of selecting individuals, 'rw', 'st' or custom function, default = 'rw'
        'rw': Roulette Wheel Selection
        'st': Stochastic Tournament Selection
        custom function: the function needs two parameters, the first is a two-dimensional darray, and the second is the fitness function
    crossover: string or function, callable, the method of crossing over, 'uniform', 'point' or custom function, default = 'uniform'
        'uniform': Uniform Crossover
        'point': Multi-point Crossover
        custom function: the function can only set one parameter with population (2-D ndarray) as input
    begin: int, callable, beginning number of the random sequence, it's used only when random_type='randint'
    end: int, callable, end number of the random sequence, it's used only when random_type='randint'
    random_state: int, callable, random seed
    select_rate: float, callable, selection rate( survival rate), default=0.3
    mutate_rate: float, callable, variation rate, default=0.05
    crossover_rate: float, callable, gene crossover probability, default=0.5
    repeat: bool, callable, whether sequence elements are allowed to repeat, default=True, only works when random_type='randint' and crossover is not a custom function

    Atrributes
    ----------
    population: int, the population size
    iteration: int, the times of iterations( the times of natural selection)
    random_type: str, the method of generating random number
    select: str, the method of selecting individuals
    crossover: str, the method of crossing over
    dom: 2-D ndarray, the dominance
    dom_fitness: 1-D ndarray, the fitness of the dominance
    '''
    def __init__(self, population=1000, iteration=100, random_type='random',
                 select='rw', crossover='uniform', begin=None, end=None,
                 random_state=None, select_rate=0.3, mutate_rate=0.05,
                 crossover_rate=0.5, repeat=True):
        self.population = population
        self.iteration = iteration
        self.random_type = random_type
        self.select = select
        self.crossover = crossover
        self.__begin = begin
        self.__end = end
        self.__select_rate = select_rate
        self.__mutate_rate = mutate_rate
        self.__crossover_rate = crossover_rate
        self.__repeat = repeat
        np.random.seed(random_state)
    
    
    def __mutate_func(self, length, populations_matrix):
        '''
        完成对新一代的变异
        '''
        mutate_matrix = np.random.rand(self.population, length) - self.__mutate_rate
        mutate_matrix = np.argwhere(mutate_matrix <= 0)

        if self.random_type == 'random':
            for i in mutate_matrix:
                populations_matrix[i[0], i[1]] = random.random()
        
        elif self.random_type == 'randint':
            if self.__repeat:
                for i in mutate_matrix:
                    populations_matrix[i[0], i[1]] = random.randint(self.__begin, self.__end)
                
            else:
                for i in mutate_matrix:
                    random_value = random.randint(self.__begin, self.__end)
                    index = np.where(populations_matrix[i[0]]==random_value)[0]
                    if len(index) == 0:
                        populations_matrix[i[0], i[1]] = random_value
                    else:
                        populations_matrix[i[0], index[0]] = populations_matrix[i[0], i[1]]
                        populations_matrix[i[0], i[1]] = random_value
        
        return populations_matrix
    
    
    def __repeat_adjust(self, child_individual_1, child_individual_2, random_loc_list):
        '''
        调整序列使得序列不存在重复元素，仅在crossover不是自定义函数时有效果
        '''
        mask = np.ones(child_individual_1.shape[0], bool)
        mask[random_loc_list] = False
        d_loc_list = np.where(mask)[0]
        child_11 = child_individual_1[random_loc_list].copy()    # 被交叉互换的片段
        child_12 = child_individual_1[d_loc_list].copy()         # 未被交叉互换的片段
        child_21 = child_individual_2[random_loc_list].copy()    # 被交叉互换的片段
        child_22 = child_individual_2[d_loc_list].copy()         # 未被交叉互换的片段

        for i in range(len(child_12)):
            while child_12[i] in child_11:
                index = np.where(child_11 == child_12[i])[0][0]
                child_12[i] = child_21[index]
            child_individual_1[d_loc_list] = child_12
        
        for i in range(len(child_22)):
            while child_22[i] in child_21:
                index = np.where(child_21 == child_22[i])[0][0]
                child_22[i] = child_11[index]
            child_individual_2[d_loc_list] = child_22
        
        return child_individual_1, child_individual_2
    
    
    def __crossover(self, num_point, length, parent_matrix, func_type):
        '''
        交叉
        '''
        if num_point:
            if num_point > length:
                raise Exception("'num_point' should be less than 'length'.")
        
        num_of_parents = len(parent_matrix)
        child_matrix = []
        child_population = self.population - num_of_parents
        
        while len(child_matrix) < child_population:
            while True:
                random_num_1 = random.randint(0, num_of_parents - 1)
                random_num_2 = random.randint(0, num_of_parents - 1)
                if random_num_1 != random_num_2:
                    break
            child_individual_1 = parent_matrix[random_num_1]
            child_individual_2 = parent_matrix[random_num_2]
            
            child_individual_1, child_individual_2, random_loc_list = func_type(self, num_point, length, child_individual_1, child_individual_2)
            if not self.__repeat:
                child_individual_1, child_individual_2 = GA.__repeat_adjust(self, child_individual_1, child_individual_2, random_loc_list)
            
            child_matrix.append(child_individual_1)
            child_matrix.append(child_individual_2)
        child_matrix = np.array(child_matrix)
        child_matrix = np.concatenate([parent_matrix, child_matrix])
        return child_matrix
    

    def __point_crossover(self, num_point, length, child_individual_1, child_individual_2):
        '''
        多点交叉
        
        
        Multi-point Crossover
        '''
        random_loc_list = []
        for j in range(num_point):
            while True:
                random_loc = random.randint(0, length - 1)
                if random_loc not in random_loc_list:
                    random_loc_list.append(random_loc)
                    break
            medium_gene = child_individual_1[random_loc]
            child_individual_1[random_loc] = child_individual_2[random_loc]
            child_individual_2[random_loc] = medium_gene
        return child_individual_1, child_individual_2, random_loc_list
    
    
    def __uniform_crossover(self, num_point, length, child_individual_1, child_individual_2):
        '''
        均匀交叉
        
        
        Uniform Crossover
        '''
        random_loc_list = []
        for j in range(length):
            crossover_possibility = random.random()
            if crossover_possibility <= self.__crossover_rate:
                medium_gene = child_individual_1[j]
                child_individual_1[j] = child_individual_2[j]
                child_individual_2[j] = medium_gene
                random_loc_list.append(j)
        return child_individual_1, child_individual_2, random_loc_list
    
    
    def __st(self, parent_matrix, fitness, num_dead):
        '''
        随机竞争选择
        num_dead: 整型，死亡的数量
        
        
        Stochastic Tournament
        num_dead: int, the number of dead individual
        '''
        num_of_parents = len(parent_matrix)
        
        for i in range(num_dead):
            while True:
                random_num_1 = random.randint(0, num_of_parents - i - 1)
                random_num_2 = random.randint(0, num_of_parents - i - 1)
                if random_num_1 != random_num_2:
                    break
            if fitness(parent_matrix[random_num_1]) <= fitness(parent_matrix[random_num_2]):
                parent_matrix = np.delete(parent_matrix, random_num_1, axis=0)
            else:
                parent_matrix = np.delete(parent_matrix, random_num_2, axis=0)
        return parent_matrix
    
    
    def __rw(self, parent_matrix, fitness, num_alive):
        '''
        轮盘赌选择
        
        
        Roulette Wheel Selection
        '''
        fitness_list = []
        child_matrix = []
        
        for parent_individual in parent_matrix:
            fitness_list.append(fitness(parent_individual))
        max_fitness = max(fitness_list)
        
        while len(child_matrix) < num_alive:
            ind = np.random.randint(0, len(parent_matrix))
            a = np.random.rand()
            if np.random.rand() > fitness_list[ind] / max_fitness or\
                parent_matrix[ind].tolist() in np.array(child_matrix).tolist():
                pass
            else:
                child_matrix.append(parent_matrix[ind])
        parent_matrix = np.array(child_matrix)
        return parent_matrix
    
    
    def fit(self, length, fitness):
        '''
        进行遗传算法模拟
        
        参数
        ----
        length：整型，染色体长度
        fitness：函数类型，适应度函数
        
        
        Simulate
        
        Parameters
        ----------
        length: int, the length of chromosome
        fitness: funtion, fitness function
        '''
        self.__fitness = fitness
        if self.random_type == 'random':
            parent_matrix = np.random.rand(self.population, length)
        elif self.random_type == 'randint':
            if not self.__repeat:
                parent_matrix = []
                for i in range(self.population):
                    parent_matrix.append(random.sample(range(self.__begin, self.__end+1), length))
                parent_matrix = np.array(parent_matrix)
            
            else:
                parent_matrix = np.random.randint(self.__begin, self.__end, size=(self.population, length))
        else:
            raise Exception("'random_type' must be one of 'random' and 'randint'.")
        
        num_alive = self.population * self.__select_rate
        num_dead = int(self.population - num_alive)
        num_point = int(length * self.__crossover_rate)
        
        for i in range(self.iteration - 1):
            # 选择
            if self.select == 'rw':
                parent_matrix = GA.__rw(self, parent_matrix, fitness, num_alive)
            elif self.select == 'st':
                parent_matrix = GA.__st(self, parent_matrix, fitness, num_dead)
            elif type(self.select).__name__ == 'function':
                parent_matrix == self.select(parent_matrix, fitness)
            
            # 交叉互换
            if self.crossover == 'uniform':
                parent_matrix = GA.__crossover(self, None, length,
                                               parent_matrix, GA.__uniform_crossover)
            elif self.crossover == 'point':
                parent_matrix = GA.__crossover(self, num_point, length,
                                               parent_matrix, GA.__point_crossover)
            
            elif type(self.crossover).__name__ == 'function':
                parent_matrix = self.crossover(parent_matrix)
            
            # 变异
            parent_matrix = GA.__mutate_func(self, length, parent_matrix)
        
        if self.select == 'rw':
            parent_matrix = GA.__rw(self, parent_matrix, fitness, num_alive)
        elif self.select == 'st':
            parent_matrix = GA.__st(self, parent_matrix, fitness, num_dead)
        elif type(self.select).__name__ == 'function':
            parent_matrix == self.select(parent_matrix, fitness)
        
        self.dom = parent_matrix
        
        dominance_fitness = []
        for individual in parent_matrix:
            dominance_fitness.append(fitness(individual))
        self.dom_fitness = np.array(dominance_fitness)


class SA():
    '''
    模拟退火算法
    默认以评估函数select的最小值为最优值
    
    参数
    ----
    anneal：浮点数类型或函数类型，可选，退火方法，若为浮点数，则按T = anneal * T退火，默认为0.9
    step：浮点数类型，可选，步长倍率，每次生成的步长为step乘一个属于(-1, 1)的随机数，默认为1
    n：整型，可选，等温时迭代次数，默认为10
    random_state：整型，可选，随机种子
    
    属性
    ----
    solution：ndarray，最优解
    trace：ndarray，迭代过程中的自变量点
    value：浮点数类型，最优解的函数值
    
    
    Simulated Annealing Algorithm
    By default, the minimum value of the evaluation function 'select' is the optimal value
    
    Parameters
    ----------
    anneal: float or function, callable, annealing method, if type is float, it will be annealed with T = anneal * T, default=0.9
    step: float, callable, step, each generated step length = step * a random number belonging to (-1, 1), default=1
    n: int, callable, isothermal iterations, default=10
    random_state: int, callable, random seed
    
    Attributes
    ----------
    solution: ndarray, optimal solution
    trace: ndarray, independent variable points in the iterative process
    value: float, function value of optimal solution
    '''
    def __init__(self, anneal=0.9, step=1, n=10, random_state=None):
        self.__anneal = anneal
        self.__step = step
        self.__n = n
        np.random.seed(random_state)
    
    
    def fit(self, select, T, T0, initial):
        '''
        参数
        ----
        select：函数，评估函数
        T：浮点数类型，初始温度
        T0：浮点数类型，退火温度
        initial：数或数组，初始解，select函数的输入值
        
        
        Parameters
        ----------
        select: function, evaluation function
        T: float, initial temperature
        T0: float, annealing temperature
        initial: num or array, initial solution, the input value of select function
        '''
        initial = np.array(initial, dtype=np.float)
        self.trace = [initial]
        while T > T0:
            for i in range(self.__n):
                random_x = 2 * self.__step * (np.random.rand(*initial.shape) - 0.5)
                initial_copy = initial.copy()
                initial_copy += random_x
                if select(initial_copy) < select(initial):
                    initial = initial_copy
                    self.trace.append(initial)
                else:
                    pro = random.random()
                    if pro < np.e**(-(select(initial_copy) - select(initial)) / T):
                        initial = initial_copy
                        self.trace.append(initial)
            
            if type(self.__anneal).__name__ == 'float':
                T *= self.__anneal
            elif type(self.__anneal).__name__ == 'function':
                T = self.__anneal(T)
            else:
                raise Exception("Type of 'anneal' must be one of 'float' and 'function'.")
        
        self.solution = initial
        self.trace = np.array(self.trace)
        self.value = select(initial)


class GD():
    '''
    梯度下降法
    沿函数负梯度方向逐步下降进而得到函数的最优解，最优解默认为最小值
    
    参数
    ----
    ytol：浮点数类型，可选，连续两次迭代的函数值小于ytol时即停止迭代，默认为0.1
    step：浮点数类型，可选，步长倍率，每次生成的步长为step * 负梯度，默认为0.1
    acc：浮点数类型，可选，求导精度，默认为0.1
    
    属性
    ----
    solution：浮点数类型，最优解
    trace：ndarray，迭代过程中的自变量点
    value：浮点数类型，最优解的函数值
    
    
    Gradient Descent
    The optimal solution of the function is obtained by gradually decreasing along the negative gradient direction
    The optimal solution is the minimum value by default
    
    Parameters
    ----------
    ytol: float, callable, when △f of two successive iterations is less than ytol, the iteration will stop, default=0.1
    step: float, callable, step, each generated step length = - step * gradient, default=1
    acc: float, callable, accuracy of derivation, default=0.1
    
    Attributes
    ----------
    solution: float, optimal solution
    trace: ndarray, independent variable points in the iterative process
    value: float, function value of optimal solution
    '''
    def __init__(self, ytol=0.1, step=0.1, acc=0.1):
        self.ytol = ytol
        self.step = step
        self.acc = acc
    
    
    def fit(self, select, initial):
        '''
        参数
        ----
        select：函数，评估函数
        initial：数或数组，初始解，select函数的输入值
        
        Parameters
        ----------
        select: function, evaluation function
        initial: num or array, initial solution, the input value of select function
        '''
        initial = np.array(initial, dtype=np.float)
        self.trace = [initial]
        f_change = float('inf')
        acc2 = 0.5 * self.acc
        
        while f_change > self.ytol:
            d_list = grad(select, initial)
            # 计算函数值变化量
            f_change = select(initial)
            initial = initial - d_list * self.step
            f_change = abs(select(initial) - f_change)
            self.trace.append(initial)
        
        self.solution = initial
        self.trace = np.array(self.trace)
        self.value = select(initial)


class GM():
    def __init__(self, acc=1):
        '''
        灰色系统模型，GM(1, 1)模型
        fit函数输入一维数组y：[y1, y2, ..., yn]
        对应的时间数组t为：[1, 2, ..., n]
        预测式：
            x(1)(t) = [x(0)(1) - b / a] * e**(- a * (t - 1)) + b / a  (t ∈ N+)
            t >= 2时：x(0)(t) = x(1)(t) - x(1)(t - 1)
            t == 1时：x(0)(t) = x(1)(t)
        
        参数
        ----
        acc：数，可选，调整级比的精度，默认为1
        
        属性
        ----
        C：数，调整级比范围时y数组的平移量（y + C）
        u：二维ndarray，列矩阵[a b].T
        predict：函数，预测函数，仅支持输入数
        
        
        Grey Model, GM(1, 1)
        The fit function inputs 1-D array y: [y1, y2, ..., yn]
        The corresponding time array t: [1, 2, ..., n]
        Prediction function:
            x(1)(t) = [x(0)(1) - b / a] * e**(- a * (t - 1)) + b / a  (t ∈ N+)
            t >= 2时：x(0)(t) = x(1)(t) - x(1)(t - 1)
            t == 1时：x(0)(t) = x(1)(t)
        
        Parameter
        ---------
        acc: num, callable, accuracy of adjusting stage ratio, default=1
        
        Attributes
        ----------
        C: num, the translation of Y array when adjusting the range of stage ratio(y + C)
        u: 2-D ndarray, column matrix [a b].T
        predict: function, prediction function, only number is supported
        '''
        self.acc=acc
    
    
    def __transform(self, y):
        # 调整级比范围
        y = np.array(y, dtype=np.float)
        n = len(y)
        y_k_1 = y[:-1]
        y_k = y[1:]
        l_k = y_k_1 / y_k
        theta_min = np.e**(-2 / (n + 1))
        theta_max = np.e**(2 / (n + 2))

        self.C = 0
        while True:
            if np.min(l_k) <= theta_min or np.max(l_k) >= theta_max:
                self.C += self.acc
                y += self.acc
                y_k_1 = y[:-1]
                y_k = y[1:]
                l_k = y_k_1 / y_k
            else:
                break
        return y
    
    
    def __generate_z1(self, y):
        # 生成y的等权邻值生成数列
        y1 = []
        for i in range(len(y)):
            y1.append(sum(y[:i+1]))
        y1 = np.array(y1, dtype=np.float)
        y1_k_1 = y1[:-1]
        y1_k = y1[1:]
        z_1 = -0.5 * y1_k_1 - 0.5 * y1_k
        return z_1
    
    
    def __generate_u(self, z1, y):
        # 求解u矩阵
        z1 = np.array([z1])
        B = np.vstack((z1, np.ones_like(z1))).T
        Y = np.array([y]).T
        u = np.linalg.lstsq(B, Y, rcond=None)[0]
        return u
    
    
    def fit(self, y):
        '''
        进行GM(1, 1)拟合
        
        参数
        ----
        y：一维数组
        
        
        Fit with GM(1, 1)
        
        Parameter
        ---------
        y: 1-D array
        '''
        y = GM.__transform(self, y)
        z1 = GM.__generate_z1(self, y)
        self.u = GM.__generate_u(self, z1, y[1:]).T[0]
        
        def predict(t):
            t = int(t)
            if t == 1:
                return self.u[1] / self.u[0] - self.C
            else:
                result = y[0] - self.u[1] / self.u[0]
                result *= (np.e**(-self.u[0] * (t - 1)) - np.e**(-self.u[0] * (t - 2)))
                return result - self.C
        
        self.predict = predict


'''
本模块用于评估和选择模型

This module is ued for estimating and selecting models
'''
import numpy as np
from matplotlib import pyplot as plt
from scipy.special import comb
plt.rcParams['axes.unicode_minus'] =False


class Error():
    '''
    误差分析
    
    参数
    ----
    func：函数类型，模型的预测函数
    
    属性
    ----
    abs_error：1-D ndarray数组，绝对误差列表
    rel_error：1-D ndarray数组，相对误差列表
    abs_sort：绝对误差从大到小的排序
    rel_sort：相对误差从大到小的排序
    mae：平均绝对误差
    mape：平均绝对百分比误差
    mse：平均平方误差
    rmse：均方根误差
    sae：绝对误差和
    sse：残差平方和
    
    
    Error Analysis
    
    Parameter
    ---------
    func: function, predicting function of models
    
    Atrributes
    ----------
    abs_error: 1-D ndarray, absolute error list
    rel_error: 1-D ndarray, relative error list
    abs_sort: list of absolute values of errors sorted from large to small
    rel_sort: list of relative values of errors sorted from large to small
    mae: mean absolute error
    mape: mean absolute percentage error
    mse: mean squared error
    rmse: root mean square error
    sae: sum of absolute error
    sse: sum of squared error
    '''
    def __init__(self, func):
        self.__func = func
    
    
    def fit(self, data, target):
        data = np.array(data)
        target = np.array(target)
        self.__num_data = len(target)
        
        predict_target = self.__func(data)
        self.abs_error = abs(target - predict_target)
        self.rel_error = abs((target - predict_target) / target)
    
    @property
    def abs_sort(self):
        error_index = np.arange(self.__num_data)
        return sorted(list(zip(self.abs_error, error_index)), reverse=True)
    
    @property
    def rel_sort(self):
        error_index = np.arange(self.__num_data)
        return sorted(list(zip(self.rel_error, error_index)), reverse=True)
    
    @property
    def sse(self):
        return sum(self.abs_error**2)
    
    @property
    def sae(self):
        return sum(self.abs_error)
    
    @property
    def mse(self):
        return sum(self.abs_error**2) / self.__num_data
    
    @property
    def mae(self):
        return sum(self.abs_error) / self.__num_data
    
    @property
    def rmse(self):
        return (sum(self.abs_error**2) / self.__num_data)**0.5
    
    @property
    def mape(self):
        return sum(self.rel_error) / self.__num_data


class Sense():
    '''
    灵敏度分析
    r = (x0, x1, x2, ..., xn)
    y = f(r)
    第i个特征在r=r0时的灵敏度：
    s(xi, r0)= (dy/dxi) * (xi/y)   (r=r0)
    
    参数
    ----
    func：函数类型，模型的预测函数，若函数需要输入列表，则列表须为ndarray
    acc：浮点数类型，可选，求导的精度，默认为0.1
    
    属性
    ----
    s_mat：由特征的灵敏度值组成的矩阵
    prediction：预测值列表
    
    
    Sensitivity Analysis
    r = (x0, x1, x2, ..., xn)
    y = f(r)
    the sensitivity of the ith feature at r=r0:
    s(xi, r0)= (dy/dxi) * (xi/y)   (r=r0)
    
    Parameters
    ----------
    func: function, predicting function of models, if the function requires a list as input, the list must be ndarray
    acc: float, callable, accuracy of derivation, default=0.1
    
    Attributes
    ----------
    s_mat: matrix combined of sensitivities of features
    prediction: list of predicted values
    '''
    def __init__(self, func, acc=0.1):
        self.__func = func
        self.acc = acc
        self.s_mat = []
        self.prediction = []
    
    
    def fit(self, x0):
        '''
        参数
        ----
        x0：数，1-D或2-D ndarray，特征的初始值，不支持批量输入
        
        
        Parameter
        ---------
        x0: num, 1-D or 2-D ndarray, initial values of features, batch input is not supported
        '''
        x0 = np.array(x0, dtype=np.float)
        acc2 = 0.5 * self.acc
        func0 = self.__func(x0)
        self.prediction.append(func0)
        s_list = []
        
        if len(x0.shape) == 0:
            x0 += acc2
            func1 = self.__func(x0)
            x0 -= self.acc
            func2 = self.__func(x0)
            s_list = (func1 - func2) / (self.acc * func0) * x0
        
        elif len(x0.shape) == 1:
            for i in range(x0.shape[0]):
                x0[i] += acc2
                func1 = self.__func(x0)
                x0[i] -= self.acc
                func2 = self.__func(x0)
                de = (func1 - func2) / (self.acc * func0) * x0[i]
                x0[i] += acc2
                s_list.append(de)
        
        elif len(x0.shape) == 2:
            for i in range(x0.shape[1]):
                x0[0][i] += acc2
                func1 = self.__func(x0)
                x0[0][i] -= self.acc
                func2 = self.__func(x0)
                de = (np.array([func1]).T - np.array([func2]).T) / (self.acc * np.array(func0).T) * x0[0][i]
                x0[0][i] += acc2
                s_list.append(de)
            
        self.s_mat.append(s_list)
    
    
    def clr(self):
        '''
        清空s_mat和prediction
        
        
        Clear the s_mat and the prediction
        '''
        self.s_mat = []
        self.prediction = []


class SHAP_and_Shapley():
    def __transform(self, b_list):
        # 二进制列表转十进制数
        d_value = 0
        for i in range(len(b_list)):
            d_value += b_list[i] * 2**i
        return d_value
    
    
    def _pre_fit(self, data, replace):
        dim = data.shape[1]
        shap_and_shapley_values = []
        
        # 求各个特征组合的预测值
        prediction = []
        for i in range(2**dim):
            new_data = data.copy()
            loc = np.array([i >>d & 1 for d in range(dim)][::1])
            loc = np.where(loc==0)[0]
            if len(loc) != 0:
                new_data[:, loc] = replace
            prediction_part = self.predict(new_data)
            prediction.append(prediction_part)
        prediction = np.array(prediction)
        
        # 求每个特征的SHAP值
        for i in range(dim):
            loc = np.array([2**i >>d & 1 for d in range(dim)][::1])
            loc = np.where(loc==1)[0][0]
            shap_value = 0
            
            # 每一层差值的加权求和
            for j in range(2**(dim-1)):
                b_list = [j >>d & 1 for d in range(dim)][::1]
                b_list.insert(loc, 0)
                b_value = SHAP_and_Shapley.__transform(self, b_list)
                # 1 / ((sum(b_list) + 1) * comb(dim, (sum(b_list) + 1)))是权重
                shap_value += (prediction[b_value + 2**i] - prediction[b_value]) / \
                              ((sum(b_list) + 1) * comb(dim, (sum(b_list) + 1)))
            shap_and_shapley_values.append(shap_value)
        return np.array(shap_and_shapley_values)
    
    
    def __image_process(self, labels, index, top):
        if not index:
            index = np.arange(0, self.values.shape[0], dtype=np.int)
        new_values = self.values[index]
        
        # 从大到小排序
        sort_list = np.argsort(new_values)
        new_values = np.sort(new_values)
        
        if not labels:
            labels = sort_list
        else:
            labels = np.array(labels, dtype=np.str)
            labels = labels[sort_list]
        
        if not top:
            top = self.values.shape[0]
        new_values = new_values[-top:]
        labels = labels[-top:]
        
        width = new_values.shape[0] * 0.5
        fig = plt.figure(figsize=(8, width))
        ax = fig.add_subplot(1, 1, 1)
        color = []
        for i in range(new_values.shape[0]):
            if new_values[i] >= 0:
                color.append('#FFE100')
            else:
                color.append('#1E90FF')
        
        rows = ax.barh(range(new_values.shape[0]), new_values, color=color)
        for rect in rows:
            w = rect.get_width()
            if w >= 0:
                ha = 'left'
            else:
                ha = 'right'
            ax.text(w, rect.get_y()+rect.get_height()/2, w, ha=ha, va='center')
        
        ax.set_yticks(range(new_values.shape[0]))
        ax.set_yticklabels(labels)
        plt.xticks(())
    
    
    def show(self, labels=None, index=None, top=None):
        '''
        作图并显示
        
        参数
        ----
        labels：一维列表，可选，特征名称，默认为None
        index：一维列表，可选，特征索引，默认为None，表示全选
        top：整型，可选，表示显示值最高的前top个特征，默认为None，表示全选
        
        
        Display the image
        
        Parameters
        ----------
        labels: 1-D list, callable, names of features, default=None
        index: 1-D list, callable, index of features, default=None, which means select all
        top: int, callable, display "top" features with the highest values, default=None, which means select all
        '''
        SHAP_and_Shapley.__image_process(self, labels, index, top)
        plt.show()
    
    
    def savefig(self, filename, labels=None, index=None, top=None):
        '''
        作图并保存
        
        参数
        ----
        filename：字符串类型，文件名
        labels：一维列表，可选，特征名称，默认为None
        index：一维列表，可选，特征索引，默认为None，表示全选
        top：整型，可选，表示显示值最高的前top个特征，默认为None，表示全选
        
        
        Save the image
        
        Parameters
        ----------
        filename: str, file name
        labels: 1-D list, callable, names of features, default=None
        index: 1-D list, callable, index of features, default=None, which means select all
        top: int, callable, display "top" features with the highest values, default=None, which means select all
        '''
        SHAP_and_Shapley.__image_process(self, labels, index, top)
        plt.savefig(filename)


class SHAP(SHAP_and_Shapley):
    def __init__(self, predict):
        '''
        计算局部点在模型中的SHAP值
        
        参数
        ----
        predict：函数类型，模型的预测函数
        
        属性
        ----
        values：一维ndarray，每个特征的SHAP值
        
        
        Calculate the SHAP of local points in the model
        
        Parameter
        ---------
        predict: function, the predict function of the model
        
        Attribute
        ---------
        values: 1-D ndarray, SHAP of each feature
        '''
        self.predict = predict
    
    
    def fit(self, data, replace=0):
        '''
        计算SHAP值
        
        参数
        ----
        data：一维数组，局部点
        replace：数或函数类型，可选，特征的替换值，函数须以np.array(data)为输入，默认为0
        
        
        Calculate the SHAP values
        
        Parameters
        ----------
        data: 1-D array, local point
        replace: num or function, callable, replacement values of features, for function, np.array(data) is the input, default=0
        '''
        data = np.array(data)
        # 确定替换值
        if type(replace).__name__ == 'function':
            replace = replace(data)
        data = np.array([data])
        
        self.values = SHAP_and_Shapley._pre_fit(self, data, replace).reshape(-1)


class Shapley(SHAP_and_Shapley):
    def __init__(self, predict):
        '''
        计算样本集在模型中的Shapley值
        
        参数
        ----
        predict：函数类型，模型的预测函数
        
        属性
        ----
        values：一维ndarray，每个特征的Shapley值
        
        
        Calculate the Shapley values of data set in the model
        
        Parameter
        ---------
        predict: function, the predict function of the model
        
        Attribute
        ---------
        values: 1-D ndarray, Shapley value of each feature
        '''
        self.predict = predict
    
    
    def fit(self, data, replace=0):
        '''
        计算Shapley
        
        参数
        ----
        data：二维数组，数据集
        replace：数或函数类型，可选，特征的替换值，函数须以np.array(data)为输入，默认为0
        
        
        Calculate the Shapley
        
        Parameters
        ----------
        data: 2-D array, data set
        replace: num or function, callable, replacement values of features, for function, np.array(data) is the input, default=0
        '''
        data = np.array(data)
        # 确定替换值
        if type(replace).__name__ == 'function':
            replace = replace(data)
        
        self.values = SHAP_and_Shapley._pre_fit(self, data, replace).T
        self.values = self.values.mean(axis=0)


'''
本模块用于部分方程求解运算

This module is used for solving equations
'''
import numpy as np
from matplotlib import pyplot as plt
from scipy.interpolate import interp1d


class Scatter():
    '''
    求解散点方程
    求解由一系列二维散点组成的曲线的交点
    
    参数
    ----
    x：一维或二维列表，仅支持以数作为自变量，不支持以列表作为自变量
    y：一维或二维列表
    acc：浮点数类型，可选，插值精度，默认0.1
    kind：浮点数类型或整型，可选，将插值类型指定为字符串('linear'、'nearest'、'zero'、'slinear'、'squardic'、'previous'、'next'，其中'zero'、'slinear'、'squared'和'cubic'表示零阶、一阶、二阶或三阶样条曲线插值；'previous'和'next'只返回点的上一个或下一个值)或作为一个整数指定要使用的样条曲线插值器的顺序。默认为'linear'
    xtol：浮点数类型，可选，横坐标误差，默认0.1
    ytol：浮点数类型，可选，纵坐标误差，默认0.01
    
    属性
    ----
    root：列表形式，解
    value：列表形式，解对应的函数值
    
    
    Solving the Equation of Dissolution Point
    Solving the intersection point of a curve composed of a series of two-dimensional scattered points
    
    Parameters
    ----------
    x: 1-D or 2-D list, only numbers are supported as arguments, lists are not
    y: 1-D or 2-D list
    acc: float, callable, interpolation accuracy, default=0.1
    kind: float or int, callable, specifies the kind of interpolation as a string ('linear', 'nearest', 'zero', 'slinear', 'quadratic', 'cubic', 'previous', 'next', where 'zero', 'slinear', 'quadratic' and 'cubic' refer to a spline interpolation of zeroth, first, second or third order; 'previous' and 'next' simply return the previous or next value of the point) or as an integer specifying the order of the spline interpolator to use. default='linear'
    xtol: float, callable, abscissa error, default=0.1
    ytol: float, callable, ordinate error, default=0.01
    
    Attributes
    ----------
    root: list, solution
    value: list, the function values of the solutions
    '''
    def __init__(self, x, y, acc=0.1, kind='linear',xtol=0.1, ytol=0.01):
        self.x = x
        self.y = y
        
        #判断输入列表维度
        try:
            x[0][0]
        except (TypeError, IndexError):
            self.x = [x]
            self.y = [y]
            zero = True
        
        min_list = []
        max_list = []
        for i in self.x:
            min_list.append(min(i))
            max_list.append(max(i))
        self.__min_x = max(min_list)
        self.__max_x = min(max_list)
        self.x_new = np.arange(self.__min_x, self.__max_x, acc)
        
        func_list = []
        self.y_new = []
        for i in range(len(self.x)):
            func = interp1d(self.x[i], self.y[i], kind=kind)
            func_list.append(func)
            self.y_new.append(func(self.x_new))
        
        if zero:
            self.y_new.append(np.zeros(len(self.x_new)))
        
        self.y_new = np.array(self.y_new)
        self.root = []
        self.value = []
        
        #构造最小范围的散点
        x_scatter = []
        y_scatter = []
        for i in range(len(self.x)):
            x_scatter.append([])
            y_scatter.append([])
            for j in range(len(self.x[i])):
                if self.__min_x <= self.x[i][j] <= self.__max_x:
                    x_scatter[i].append(self.x[i][j])
                    y_scatter[i].append(self.y[i][j])
        self.x = x_scatter
        self.y = y_scatter
        
        x_buffer = []
        y_buffer = []
        for j in range(self.y_new.shape[1]):
            delta = []
            for i in range(self.y_new.shape[0]):
                delta.append(self.y_new[i, j])
            mean_delta = np.mean(delta)
            delta = max(delta) - min(delta)
            if delta <= ytol:
                x_buffer.append(self.x_new[j])
                y_buffer.append(mean_delta)
                if x_buffer[-1] - x_buffer[0] > xtol:
                    self.root.append(np.mean(x_buffer[:-1]))
                    x_buffer = [x_buffer[-1]]
                    self.value.append(np.mean(y_buffer[:-1]))
                    y_buffer = [y_buffer[-1]]
                
            if j == self.y_new.shape[1] - 1:
                self.root.append(np.mean(x_buffer))
                self.value.append(np.mean(y_buffer))
    
    
    def __image_process(self, scatter):
        for i in range(len(self.x)):
            plt.plot(self.x_new, self.y_new[i])
            if scatter:
                plt.plot(self.x[i], self.y[i], 'o')
        plt.plot(self.root, self.value, 'ro')
        
        
    def show(self, scatter=False):
        '''
        作图并显示
        
        参数
        ---
        scatter：布尔类型，True表示显示原散点，False表示不显示
        
        
        Display the image
        
        Parameter
        ---------
        scatter: bool, True means display the origin scatter, False means not
        '''
        Scatter.__image_process(self, scatter=scatter)
        plt.show()
    
    
    def savefig(self, filename, scatter=False):
        '''
        作图并保存
        
        参数
        ----
        filename：字符串类型，保存的文件名
        scatter：布尔类型，True表示显示原散点，False表示不显示
        
        
        Save the image
        
        Parameters
        ----------
        filename: str, file name
        scatter: bool, True means display the origin scatter, False means not
        '''
        Scatter.__image_process(self, scatter=scatter)
        plt.savefig(filename)


def linear_e(A, b):
    '''
    求解线性方程组Ax = b
    
    参数
    ----
    A: 二维列表，系数矩阵
    b：二维列表
    
    返回
    ----
    二维ndarray，方程的解
    
    
    Solve linear system of equations Ax = b
    
    Parameters
    ----------
    A: 2-D list, coefficient matrix
    b: 2-D list
    
    Return
    ------
    2-D ndarray, the solution of the equations
    '''
    A = np.array(A)
    b = np.array(b)
    A = np.linalg.pinv(A)
    return np.dot(A, b)